# -*- coding: utf-8 -*-
""" A simple module to store sites from fusion tables """
import ee
import csv
import requests


class Site(object):
    """ Site Class to store sites related to fusion tables """
    def __init__(self, name=None, id_ft=None, id_fld=None, name_fld=None,
                 n_feat=0):
        """
        :param name: name of the site
        :type name: str
        :param id_ft: Fusion Table ID
        :type id_ft: str
        :param id_fld: Name of the Fusion Table Field that contains the id for
            each individual site
        :type id_fld: str
        :param name_fld: Name of the Fusion Table Field that contains the name
            for each individual site
        :type name_fld: str
        :param n_feat: Number of features that contain the Fusion Table
        :type n_feat: int
        """
        self.name = name
        self.id_ft = id_ft
        self.id_fld = id_fld
        self.name_fld = name_fld
        self.n_feat = n_feat

    @property
    def ft(self):
        if self.id_ft:
            return ee.FeatureCollection("ft:"+self.id_ft)
        else:
            return None

    def filter_id(self, id):
        """ Filters the fusion table by the given id

        :param id: id to filter
        :type id: int
        :return: ee.Feature, region (as a list of lists)
        :rtype: tuple
        """
        try:
            place = self.ft.filterMetadata(self.id_fld, "equals", id)
            place = ee.Feature(place.first())
            place = place.set("origin", self.name, "id", id)

            try:
                region = place.geometry().bounds().getInfo()['coordinates'][0]
            except AttributeError as ae:
                print(ae)
                region = place.getInfo()['coordinates'][0]
            except Exception as e:
                print(e)
                return None, None

            return place, region
        except Exception as e:
            # print "Hubo un error al filtrar el ID"
            print(e)
            return None, None


def from_csv(file, name="name", id_ft="id_ft", id_fld="id_fld",
             name_fld=None):
    """ Generates a dictionary of Sites from a csv file.

    :param name:
    :param id_ft:
    :param id_fld:
    :param name_fld:
    """
    sites = []
    with open(file) as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            params = [row[name], row[id_ft], row[id_fld]]
            params = params.append(row[name_fld]) if name_fld else params
            site = (row[name], Site(*params))
            sites.append(site)

    return dict(sites)


def from_gsheet(url, sheet, name=None, id_ft=None, id_fld=None, name_fld=None):
    """ Generates a dictionary of Sites from a Google SpreadSheet. It must be
    Public and shared to anyone. Doesn't use any API

    :param url:
    :param sheet:
    :param name: name of field that holds the name of the site
    :param id_ft: name of field that holds the id of the fusion table
    :param id_fld: name of field that holds the ID of the site
    :param name_fld:
    """
    content = requests.get(url)
    json = content.json()
    sheet = json[sheet]
    sites = []

    for n, row in enumerate(sheet):
        if n == 0: continue
        if row[name] == "": continue
        # params = [row[name], row[id_ft], row[id_fld]]
        # print params
        # params = params.append(row[name_fld]) if name_fld else params
        site = (row[name], Site(name=row[name], id_ft=row[id_ft],
                                id_fld=row[id_fld], name_fld=row[name_fld]))
        sites.append(site)

    return dict(sites)



# -*- coding: utf-8 -*-
""" Date module for Gee Bap """
import ee


class Date(object):
    """ Class holding some custom methods too add a 'date' band in the 'Best
    Available Pixel' compostite.

    Agrega una band de name *fecha* de tipo *Uint16* que toma de la
    propiedad *system:time_start* y computa la cantidad de dias que
    transcurrieron a partir del 1970-01-01.
    Agrega a la imagen una propiedad con el id de la imagen llamada 'id_img'

    :metodos:
    :mapfecha: estatico para utilizar con ee.ImageCollection.map()
    """
    oneday_local = 86400000  # milisegundos
    oneday = ee.Number(oneday_local)

    def __init__(self):
        ''' This Class doesn't initialize '''
        pass

    @staticmethod
    def map(name="date"):
        """
        :param name: name for the new band
        :type name: str
        """
        def wrap(img):
            dateadq = img.date()  # ee.Date
            days_since_70 = ee.Date(dateadq).millis().divide(Date.oneday)  # days since 1970
            dateimg = ee.Image(days_since_70).select([0], [name]).toUint16()
            final = img.addBands(dateimg).set(name, days_since_70.toInt())
            # return functions.pass_date(img, final)
            # return tools.passProperty(img, final, 'system:time_start')
            return final.copyProperties(img, ['system:time_start'])
        return wrap

    @staticmethod
    def local(date):
        """ Number of days since the beggining (1970-01-01)
        Dada una fecha obtiene la cantidad de dias desde el comienzo
        de las fechas (01-01-1970)

        :param date: date (yyyy-MM-dd)
        :type date: str
        :return: days since the beggining
        :rtype: float
        """
        d = ee.Date(date)
        mili = d.millis().getInfo()
        return float(mili / Date.oneday_local)

    @staticmethod
    def get(date, unit="days"):
        """ get the date (ee.Date) of the given value in 'unit'.
        Currentrly ONLY process 'days', so:

        `date.Date.get(365) = '1971-01-01T00:00:00`

        :param date: the value to transform
        :type date: int
        :param unit: date's unit (currently ONLY 'days')
        :return: date corresponding to the given value
        :rtype: ee.Date
        """
        if unit == "days":
            mili = ee.Number(date).multiply(Date.oneday)
            d = ee.Date(mili)
            # dstr = d.format()
        return d



# -*- coding: utf-8 -*-
""" Main module holding the Bap Class and its methods """

from geetools import collection, tools
from . import scores, priority, functions, utils, __version__
import ee
import json


class Bap(object):
    def __init__(self, season, range=(0, 0), colgroup=None, scores=None,
                 masks=None, filters=None, target_collection=None, brdf=False,
                 harmonize=True, **kwargs):
        self.range = range
        self.scores = scores
        self.masks = masks
        self.filters = filters
        self.season = season
        self.colgroup = colgroup
        self.brdf = brdf
        self.harmonize = harmonize

        if target_collection is None:
            target_collection = collection.Landsat8SR()
        self.target_collection = target_collection

        self.score_name = kwargs.get('score_name', 'score')

        # Band names in case user needs different names
        self.bandname_col_id = kwargs.get('bandname_col_id', 'col_id')
        self.bandname_date = kwargs.get('bandname_date', 'date')

    @property
    def score_names(self):
        if self.scores:
            punt = [p.name for p in self.scores]
            return functions.replace_duplicate(punt)
        else:
            return []

    @property
    def max_score(self):
        """ gets the maximum score it can get """
        maxpunt = 0
        for score in self.scores:
            maxpunt += score.max
        return maxpunt

    def year_range(self, year):
        try:
            i = year - abs(self.range[0])
            f = year + abs(self.range[1]) + 1

            return range(i, f)
        except:
            return None

    def time_start(self, year):
        """ Get time start property """
        return ee.Date('{}-{}-{}'.format(year, 1, 1))

    def make_proxy(self, image, collection, year):
        """ Make a proxy collection """

        size = collection.size()

        # unmask all bands
        unmasked = image.unmask()

        proxy_date = ee.Date('{}-01-01'.format(year))

        bands = image.bandNames()
        empty = tools.image.empty(0, bands)
        proxy = unmasked.where(unmasked, empty)

        proxy = proxy.set('system:time_start', proxy_date.millis())

        proxy_col = ee.ImageCollection.fromImages([proxy])

        return ee.ImageCollection(ee.Algorithms.If(size.gt(0),
                                                   collection,
                                                   proxy_col))


    def compute_scores(self, year, site, indices=None, **kwargs):
        """ Add scores and merge collections

        :param add_individual_scores: adds the individual scores to the images
        :type add_individual_scores: bool
        :param buffer: make a buffer before cutting to the given site
        :type buffer: float
        """
        add_individual_scores = kwargs.get('add_individual_scores', False)
        buffer = kwargs.get('buffer', None)

        # list to 'collect' collections
        all_collections = ee.List([])

        # TODO: get common bands for col of all years
        if self.colgroup is None:
            colgroup = priority.SeasonPriority(year).colgroup
            all_col = []
            for year in self.year_range(year):
                _colgroup = priority.SeasonPriority(year).colgroup
                for col in _colgroup.collections:
                    all_col.append(col)
        else:
            all_col = self.colgroup.collections
            colgroup = self.colgroup

        common_bands = collection.getCommonBands(*all_col, match='name')

        # add col_id to common bands
        common_bands.append(self.bandname_col_id)

        # add date band to common bands
        common_bands.append(self.bandname_date)

        # add score names if 'add_individual_scores'
        if add_individual_scores:
            for score_name in self.score_names:
                common_bands.append(score_name)

        # add indices to common bands
        if indices:
            for i in indices:
                common_bands.append(i)

        # add score band to common bands
        common_bands.append(self.score_name)

        # create an empty score band in case no score is parsed
        empty_score = ee.Image.constant(0).rename(self.score_name).toUint8()

        # List to store all used images
        used_images = dict()

        for col in colgroup.collections:
            col_ee_bounds = col.collection

            # Filter bounds
            if isinstance(site, ee.Feature): site = site.geometry()
            col_ee_bounds = col_ee_bounds.filterBounds(site)

            # Collection ID
            col_id = functions.get_col_id(col)
            col_id_img = functions.get_col_id_image(col)

            for year in self.year_range(year):
                daterange = self.season.add_year(year)

                # filter date
                col_ee = col_ee_bounds.filterDate(daterange.start(),
                                                  daterange.end())

                # some filters
                if self.filters:
                    for filt in self.filters:
                        if filt.name in ['CloudCover']:
                            col_ee = filt.apply(col_ee, col=col)

                # BRDF
                if self.brdf:
                    if 'brdf' in col.algorithms.keys():
                        col_ee = col_ee.map(lambda img: col.brdf(img))

                # Proxy in case size == 0
                col_ee = self.make_proxy(col.collection.first(), col_ee, year)

                # clip with site
                if buffer is not None:
                    site = site.buffer(buffer)
                col_ee = col_ee.map(lambda img: img.clip(site))

                # Add year as a property (YEAR_BAP)
                col_ee = col_ee.map(lambda img: img.set('YEAR_BAP', year))

                # Catch SLC off
                slcoff = False
                if col.spacecraft == 'LANDSAT' and col.number == 7:
                    if year in priority.SeasonPriority.l7_slc_off:
                        # Convert masked values to zero
                        col_ee = col_ee.map(lambda img: functions.unmask_slc_off(img))
                        slcoff = True

                # Apply masks
                if self.masks:
                    for mask in self.masks:
                        col_ee = mask.map(col_ee, col=col)

                # Rename
                col_ee = col_ee.map(lambda img: col.rename(img))

                # Rescale
                col_ee = col_ee.map(
                    lambda img: collection.rescale(
                        img, col, self.target_collection, renamed=True))

                # Indices
                if indices:
                    for i in indices:
                        f = getattr(col, i)
                        def addindex(img):
                            ind = f(img, renamed=True)
                            return img.addBands(ind)
                        col_ee = col_ee.map(addindex)

                # Apply scores
                if self.scores:
                    for score in self.scores:
                        zero = False if slcoff and isinstance(score, (scores.MaskPercent, scores.MaskPercentKernel)) else True
                        col_ee = score._map(
                            col_ee,
                            col=col,
                            year=year,
                            colEE=col_ee,
                            geom=site,
                            include_zero=zero)

                # Mask all bands with mask
                col_ee = col_ee.map(lambda img: img.updateMask(img.select([0]).mask()))

                # Get an image before the filter to catch all bands for proxy image
                col_ee_image = col_ee.first()

                # Filter Mask Cover
                if self.filters:
                    for filt in self.filters:
                        if filt.name in ['MaskCover']:
                            col_ee = filt.apply(col_ee)

                # col_ee = self.make_proxy(col, col_ee, year, True)
                col_ee = self.make_proxy(col_ee_image, col_ee, year)

                # Add col_id band
                # Add col_id to the image as a property
                def addBandID(img):
                    return img.addBands(col_id_img).set(
                        self.bandname_col_id.upper(), col_id)
                col_ee = col_ee.map(addBandID)

                # Add date band
                def addDateBand(img):
                    date = img.date()
                    year = date.get('year').format()

                    # Month
                    month = date.get('month')
                    month_str = month.format()
                    month = ee.String(ee.Algorithms.If(
                        month.gte(10),
                        month_str,
                        ee.String('0').cat(month_str)))

                    # Day
                    day = date.get('day')
                    day_str = day.format()
                    day = ee.String(ee.Algorithms.If(
                        day.gte(10),
                        day_str,
                        ee.String('0').cat(day_str)))

                    date_str = year.cat(month).cat(day)
                    newdate = ee.Number.parse(date_str)
                    newdate_img = ee.Image.constant(newdate) \
                        .rename(self.bandname_date).toUint32()
                    return img.addBands(newdate_img)
                col_ee = col_ee.map(addDateBand)

                # Harmonize
                if self.harmonize:
                    # get max value for the needed bands

                    if 'harmonize' in col.algorithms.keys():
                        col_ee = col_ee.map(
                            lambda img: col.harmonize(img, renamed=True))

                # store used images
                # Property name for storing images as properties
                prop_name = 'BAP_IMAGES_COLID_{}_YEAR_{}'.format(col_id, year)
                # store used images
                imlist = ee.List(col_ee.toList(col_ee.size()).map(
                    lambda img:
                    ee.String(col.id).cat('/').cat(ee.Image(img).id())))
                used_images[prop_name] = imlist

                col_ee_list = col_ee.toList(col_ee.size())
                all_collections = all_collections.add(col_ee_list).flatten()

        all_collection = ee.ImageCollection.fromImages(all_collections)

        # Compute final score
        if self.scores:
            def compute_score(img):
                score = img.select(self.score_names).reduce('sum') \
                    .rename('score').toFloat()
                return img.addBands(score)
        else:
            def compute_score(img):
                return img.addBands(empty_score)

        all_collection = all_collection.map(compute_score)

        # Select common bands
        # all_collection = functions.select_match(all_collection)
        final_collection = all_collection.map(
            lambda img: img.select(common_bands))

        self._used_images = used_images

        return final_collection

    def build_composite_best(self, year, site, indices=None, **kwargs):
        """ Build the a composite with best score

        :param add_individual_scores: adds the individual scores to the images
        :type add_individual_scores: bool
        :param buffer: make a buffer before cutting to the given site
        :type buffer: float
        """
        # TODO: pass properties
        col = self.compute_scores(year, site, indices, **kwargs)
        mosaic = col.qualityMosaic(self.score_name)

        return self._set_properties(mosaic, year, col)

    def build_composite_reduced(self, year, site, indices=None, **kwargs):
        """ Build the composite where

        :param add_individual_scores: adds the individual scores to the images
        :type add_individual_scores: bool
        :param buffer: make a buffer before cutting to the given site
        :type buffer: float
        """
        # TODO: pass properties
        nimages = kwargs.get('set', 5)
        reducer = kwargs.get('reducer', 'interval_mean')
        col = self.compute_scores(year, site, indices, **kwargs)
        mosaic = reduce_collection(col, nimages, reducer, self.score_name)

        return self._set_properties(mosaic, year, col)

    def _set_properties(self, mosaic, year, col):
        """ Set some BAP common properties to the given mosaic """
        # # USED IMAGES
        # used_images = self._used_images
        # for prop, value in used_images.items():
        #     mosaic = mosaic.set(prop, str(value))

        # SCORES
        bap_scores = []
        for score in self.scores:
            pattern = utils.object_init(score)
            bap_scores.append(pattern)
        mosaic = mosaic.set('BAP_SCORES', str(bap_scores))

        # MASKS
        bap_masks = []
        for mask in self.masks:
            pattern = utils.object_init(mask)
            bap_masks.append(pattern)
        mosaic = mosaic.set('BAP_MASKS', str(bap_masks))

        # FILTERS
        bap_filters = []
        for filter in self.filters:
            pattern = utils.object_init(filter)
            bap_filters.append(pattern)
        mosaic = mosaic.set('BAP_FILTERS', str(bap_filters))
        
        # DATE
        date = self.time_start(year).millis()
        mosaic = mosaic.set('system:time_start', date)

        # BAP Version
        mosaic = mosaic.set('BAP_VERSION', __version__)

        # FOOTPRINT
        geom = tools.imagecollection.mergeGeometries(col)
        mosaic = mosaic.set('system:footprint', geom)

        # Seasons
        for year in self.year_range(year):
            yearstr = ee.Number(year).format()
            daterange = self.season.add_year(year)
            start = daterange.start().format('yyyy-MM-dd')
            end = daterange.end().format('yyyy-MM-dd')
            string = start.cat(' to ').cat(end)
            propname = ee.String('BAP_SEASON_').cat(yearstr)
            mosaic = mosaic.set(propname, string)

        return mosaic

    def to_file(self, filename, path=None):
        """ Make a configuration file (JSON) to be able to reconstruct the
        object """
        import os
        def make_name(filename):
            split = filename.split('.')
            if len(split) == 1:
                filename = '{}.json'.format(filename)
            else:
                ext = split[-1]
                if ext != 'json':
                    name = '.'.join(split[:-1])
                    filename = '{}.json'.format(name)
            return filename
        filename = make_name(filename)
        path = '' if path == None else path
        path = os.path.join(os.getcwd(), path, filename)
        serial = utils.serialize(self, 'config')['config (Bap)']
        with open(path, 'w') as thefile:
            json.dump(serial, thefile, indent=2)


def load(filename, path=None):
    """ Create a Bap object using a config file """
    from . import season, masks, filters, scores
    from geetools import collection
    import os
    if path is None:
        path = os.getcwd()
    split = filename.split('.')
    if split[-1] != 'json':
        filename = '{}.json'.format(filename)
    obj = json.load(open(os.path.join(path, filename)))

    # HELPER
    def get_number(param_dict, name):
        if name != '':
            params = ['{} (int)', '{} (float)']
        else:
            params = ['{}(int)', '{}(float)']

        for param in params:
            result = param_dict.get(param.format(name))
            if result is not None:
                return result
        return None

    # SEASON
    seas = obj['season (Season)']
    start = seas['_start (SeasonDate)']['date (str)']
    end = seas['_end (SeasonDate)']['date (str)']
    season_param = season.Season(start, end)

    # RANGE
    ran = obj.get('range (tuple)') or obj.get('range (list)')
    range_param = (ran[0]['(int)'], ran[1]['(int)'])

    # MASKS
    mask_list = []
    for mask in obj.get('masks (tuple)') or obj.get('masks (list)') or []:
        mask_class = list(mask.keys())[0]
        params = mask[mask_class]
        if mask_class == '(Mask)':
            options = [opt['(str)'] for opt in params.get('options (list)') or params.get('options (tuple)')]
            mask_param = masks.Mask(options)
        elif mask_class == '(Hollstein)':
            options = [opt['(str)'] for opt in params.get('options (list)') or params.get('options (tuple)')]
            mask_param = masks.Hollstein(options)
        else:
            continue
        mask_list.append(mask_param)

    # FILTERS
    filter_list = []
    for filter in obj.get('filters (tuple)') or obj.get('filters (list)') or []:
        filter_class = list(filter.keys())[0]
        params = filter[filter_class]
        if filter_class == '(CloudCover)':
            percent = get_number(params, 'percent')
            filter_param = filters.CloudCover(percent)
        elif filter_class == 'MaskCover':
            percent = get_number(params, 'percent')
            filter_param = filters.MaskCover(percent)
        else:
            continue
        filter_list.append(filter_param)

    # COLGROUP
    colgroup = obj.get('colgroup (CollectionGroup') or obj.get('colgroup (NoneType)')
    if colgroup:
        collections = []
        collections_param = colgroup.get('collections (tuple)') or colgroup.get('collections (list)')
        for col in collections_param:
            sat = list(col.keys())[0]
            params = col[sat]
            satid = params['_id (str)']
            instance = collection.fromId(satid)
            collections.append(instance)
        colgroup_param = collection.CollectionGroup(collections)
    else:
        colgroup_param = None


    # TARGET COLLECTION
    target_collection = obj.get('target_collection (Landsat)') or \
                        obj.get('target_collection (Sentinel)')

    target_id = target_collection.get('id (str)')
    target_param = collection.fromId(target_id)

    # SCORES
    score_list = []
    for score in obj.get('scores (list)') or obj.get('scores (tuple)') or []:
        score_class = list(score.keys())[0]
        params = score[score_class]
        # RANGE OUT
        range_out_param = params.get('range_out (tuple)') or params.get('range_out (list)')
        if range_out_param:
            range_out_0 = range_out_param[0]
            range_out_1 = range_out_param[1]
            range_out = (get_number(range_out_0, ''), get_number(range_out_1, ''))
        else:
            range_out = None

        # RANGE IN
        range_in_param = params.get('range_in (tuple)') or params.get('range_in (list)')
        if range_in_param:
            range_in_0 = range_in_param[0]
            range_in_1 = range_in_param[1]
            range_in = (get_number(range_in_0, ''), get_number(range_in_1, ''))
        else:
            range_in = None

        # NAME
        name = params.get('name (str)')
        sleep = get_number(params, 'sleep')
        if score_class == '(CloudScene)':
            continue
        if score_class == '(CloudDist)':
            dmax = get_number(params, 'dmax')
            dmin = get_number(params, 'dmin')
            kernel = params.get('kernel (str)')
            units = params.get('units (str)')
            score_param = scores.CloudDist(dmin, dmax, name, kernel=kernel, units=units)
        elif score_class == '(Doy)':
            best_doy = params.get('best_doy (str)')
            doy_season_param = params.get('season (Season)')
            start = doy_season_param['_start (SeasonDate)']['date (str)']
            end = doy_season_param['_end (SeasonDate)']['date (str)']
            Season = season.Season(start, end)
            function = params.get('function (str)')
            stretch = get_number(params, 'stretch')
            score_param = scores.Doy(best_doy, Season, name, function, stretch)
        elif score_class == '(AtmosOpacity)':
            continue
        elif score_class == '(MaskPercent)':
            band = params.get('band (str)')
            maxPixels = params.get('maxPixels (int)')
            count_zeros = params.get('count_zeros (bool)')
            score_param = scores.MaskPercent(band, name, maxPixels, count_zeros)
        elif score_class == '(MaskPercentKernel)':
            kernel = params.get('kernel (str)')
            distance = get_number(params, 'distance')
            units = params.get('units (str)')
            score_param = scores.MaskPercentKernel(kernel, distance, units, name)
        elif score_class == '(Satellite)':
            ratio = get_number(params, 'ratio')
            score_param = scores.Satellite(ratio, name)
        elif score_class == '(Outliers)':
            bands = params.get('bands (tuple)') or params.get('bands (list)')
            bandlist = [band['(str)'] for band in bands]
            process = params.get('process (str)')
            dist = get_number(params, 'dist')
            score_param = scores.Outliers(bandlist, process, dist, name)
        elif score_class == '(Index)':
            index = params.get('index (str)')
            target = get_number(params, 'target')
            function = params.get('function (str)')
            stretch = get_number(params, 'stretch')
            score_param = scores.Index(index, target, name, function, stretch)
        elif score_class == '(MultiYear)':
            main_year = params.get('main_year (int)')
            my_season_param = params.get('season (Season)')
            start = my_season_param['_start (SeasonDate)']['date (str)']
            end = my_season_param['_end (SeasonDate)']['date (str)']
            Season = season.Season(start, end)
            ratio = get_number(params, 'ratio')
            function = params.get('function (str)')
            stretch = get_number(params, 'stretch')
            score_param = scores.MultiYear(main_year, Season, ratio, function, stretch, name)
        elif score_class == '(Threshold)':
            continue
        elif score_class == '(Medoid)':
            bands = params.get('bands (list)') or params.get('bands (tuple)')
            discard_zeros = params.get('discard_zeros (bool)')
            score_param = scores.Medoid(bands, discard_zeros, name)
        elif score_class == '(Brightness)':
            target = get_number(params, 'target')
            bands = params.get('bands (list)') or params.get('bands (tuple)')
            function = params.get('function (str)')
            score_param = scores.Brightness(target, bands, name, function)
        else:
            continue

        score_param.sleep = sleep
        score_param.range_out = range_out
        score_param.range_in = range_in
        score_list.append(score_param)

    # MISC
    score_name_param = obj.get('score_name (str)')
    bandname_date_param = obj.get('bandname_date (str)')
    brdf_param = obj.get('brdf (bool)')
    harmonize_param = obj.get('harmonize (bool)')
    bandname_col_id_param = obj.get('bandname_col_id (str)')

    return Bap(season_param, range_param, colgroup_param, score_list,
               mask_list, filter_list, target_param, brdf_param,
               harmonize_param, score_name=score_name_param,
               bandname_date=bandname_date_param,
               bandname_col_id=bandname_col_id_param)


def reduce_collection(collection, set=5, reducer='mean',
                      scoreband='score'):
    """ Reduce the collection and get a statistic from a set of pixels

    Transform the collection to a 2D array

    * axis 1 (horiz) -> bands
    * axis 0 (vert) -> images

    ====== ====== ===== ===== ======== =========
       \     B2    B3    B4     index    score
    ------ ------ ----- ----- -------- ---------
     img1
     img2
     img3
    ====== ====== ===== ===== ======== =========

    :param reducer: Reducer to use for the set of images. Options are:
        'mean', 'median', 'mode', 'intervalMean'(default)
    :type reducer: str || ee.Reducer
    :param collection: collection that holds the score
    :type collection: ee.ImageCollection
    :return: An image in which every pixel is the reduction of the set of
        images with best score
    :rtype: ee.Image
    """
    reducers = {'mean': ee.Reducer.mean(),
                'median': ee.Reducer.median(),
                'mode': ee.Reducer.mode(),
                'interval_mean': ee.Reducer.intervalMean(50, 90),
                'first': ee.Reducer.first(),
                }

    if reducer in reducers.keys():
        selected_reducer = reducers[reducer]
    elif isinstance(reducer, ee.Reducer):
        selected_reducer = reducer
    else:
        raise ValueError('Reducer {} not recognized'.format(reducer))

    # Convert masked pixels to 0 value
    # collection = collection.map(tools.mask2zero)
    collection = collection.map(lambda img: img.unmask())

    array = collection.toArray()

    # Axis
    bands_axis = 1
    images_axis = 0

    # band names
    bands = ee.Image(collection.first()).bandNames()

    # index of score
    score_index = bands.indexOf(scoreband)

    # get only scores band
    score = array.arraySlice(axis= bands_axis,
                             start= score_index,
                             end= score_index.add(1))

    # Sort the array (ascending) by the score band
    sorted_array = array.arraySort(score)

    # total longitud of the array image (number of images)
    longitud = sorted_array.arrayLength(0)

    # Cut the Array
    # lastvalues = arrayOrdenado.arraySlice(self.ejeImg,
    # longitud.subtract(self.set), longitud)
    lastvalues = sorted_array.arraySlice(axis=images_axis,
                                          start=longitud.subtract(set),
                                          end=longitud)

    # Cut score axis
    solopjes = lastvalues.arraySlice(axis=bands_axis,
                                     start=score_index,
                                     end= score_index.add(1))

    #### Process ####
    processed = lastvalues.arrayReduce(selected_reducer,
                                       ee.List([images_axis]))

    # Transform the array to an Image
    result_image = processed.arrayProject([bands_axis]) \
                            .arrayFlatten([bands])

    return result_image


# -*- coding: utf-8 -*-
"""
Module to implement scores in the Bap Image Composition.

Scores can be computed using a single image, for example a score for percentage
of masked pixels, or using a collection, for example a score for outliers.

Each score must have an `apply` method which first argument must be an
ImageCollection and must return the collection with score computed in each
image. This method should be an staticmethod and not depend on any external
library except Earth Engine API.

Each score must have an `compute` method which first argument must be an Image
and return the resulting score as an Image with one band. This method should be
an staticmethod and not depend on any external library except Earth Engine API.

If the `compute` method is borrowed from another package it can be passed by
overwriting it. It should be use only internally by `apply` method, and `apply`
method should be the only one used by the BAP process.

The `map` method is designed to be used in the BAP method only, it takes an
ImageCollections as first argument and can use the following keyword arguments
(kwargs):

- col: a satcol.Collection object instance
- year: the given year
- geom: a geometry
- any other keyword argument

"""
import ee

from . import priority
from . import season as season_module
from geetools import tools, composite

from .expressions import Expression
from abc import ABCMeta, abstractmethod
from .regdec import *

from uuid import uuid4

__all__ = []
factory = {}

KERNELS_DISTANCE = {
    "euclidean": ee.Kernel.euclidean,
    "manhattan": ee.Kernel.manhattan,
    "chebyshev": ee.Kernel.chebyshev
}

KERNELS_BOOL = {
    "circle": ee.Kernel.circle,
    "cross": ee.Kernel.cross,
    "diamond": ee.Kernel.diamond,
    "octagon": ee.Kernel.octagon,
    "plus": ee.Kernel.plus,
    "square": ee.Kernel.square
}


class Score(object):
    ''' Abstract Base class for scores '''
    __metaclass__ = ABCMeta

    def __init__(self, name="score", range_in=None, range_out=(0, 1), sleep=0,
                 **kwargs):
        """ Abstract Base Class for scores

        :param name: score's name
        :type name: str
        :param range_in: score's range
        :type range_in: tuple
        :param sleep: time to wait until to compute the next score
        :type sleep: int
        :param formula: formula to use for computing the score
        :type formula: Expression
        :param range_out: Adjust the output to this range
        :type range_out: tuple
        """
        self.name = name
        self.range_in = range_in
        self.range_out = range_out
        self.sleep = sleep
        self._normalize = True

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, value):
         self._normalize = value

    @property
    def max(self):
        return self.range_out[1]

    @property
    def min(self):
        return self.range_out[0]

    def adjust(self):
        if self.range_out != (0, 1):
            return lambda img: tools.image.parametrize(img, (0, 1),
                                                       self.range_out,
                                                       [self.name])
        else:
            return lambda x: x

    @abstractmethod
    def map(self, collection, **kwargs):
        """ Abstract map method to use in ImageCollection.map() """
        return collection

    @staticmethod
    def compute(img, **kwargs):
        """ Compute method. This method is supposed to be the one to
        compute the score
        """
        def wrap(img):
            return img
        return wrap

    @staticmethod
    def apply(collection, **kwargs):
        return collection

    def empty(self, img):
        """ Make an empty score band. All pixels will have zero value """
        i = ee.Image.constant(0).select([0], [self.name]).toFloat()
        return img.addBands(i)

    def _map(self, collection, **kwargs):
        """ Internal map function for applying adjust """
        newcollection = self.map(collection, **kwargs)

        return newcollection.map(self.adjust())


@register(factory)
@register_all(__all__)
class CloudScene(Score):
    """ Cloud cover percent score for the whole scene. Default name for the
    resulting band will be 'score-cld-esc'.

    :param name: name of the resulting band
    :type name: str
    """
    def __init__(self, name="score-cld-scene", **kwargs):
        super(CloudScene, self).__init__(**kwargs)
        self.range_in = (0, 100)
        self.name = name

        self.formula = Expression.Exponential(rango=self.range_in,
                                              normalizar=self.normalize,
                                              **kwargs)

    @staticmethod
    def compute(img, **kwargs):
        formula = kwargs.get('formula')
        fmap = kwargs.get('fmap')
        name = kwargs.get('name')
        cloud_cover = kwargs.get('cloud_cover')

        func = formula.map(name, prop=cloud_cover, map=fmap)
        return func(img)

    @staticmethod
    def apply(collection, **kwargs):
        return collection.map(lambda img: CloudScene.compute(img, **kwargs))

    def map(self, collection, **kwargs):
        """ Map function to use in BAP

        :param col: collection
        :type col: satcol.Collection
        """
        col = kwargs.get('col')

        if col.cloud_cover:
            return collection.map(
                lambda img: self.compute(img, cloud_cover=col.cloud_cover))
        else:
            return collection.map(self.empty)


@register(factory)
@register_all(__all__)
class CloudDist(Score):
    """ Score for the distance to the nearest cloud. Default name will be
    'score-cld-dist'

    :param unit: Unit to use in the distance kernel. Options: meters,
        pixels
    :type unit: str
    :param dmax: Maximum distance to calculate the score. If the pixel is
        further than dmax, the score will be 1.
    :type dmax: int
    :param dmin: Minimum distance.
    :type dmin: int
    """
    def __init__(self, dmin=0, dmax=None, name="score-cld-dist", **kwargs):
        super(CloudDist, self).__init__(**kwargs)
        self.dmax = dmax
        self.dmin = dmin
        self.name = name
        self.range_in = (dmin, dmax)
        # Extra params
        self.sleep = kwargs.get("sleep", 10)
        self.kernel = kwargs.get("kernel", "euclidean")
        self.units = kwargs.get('units', 'meters')

    # GEE
    @property
    def dminEE(self):
        return ee.Image.constant(self.dmin)

    def kernelEE(self, radius):
        fkernel = KERNELS_DISTANCE[self.kernel]
        return fkernel(radius=radius, units=self.units)

    @staticmethod
    def compute(img, **kwargs):
        """ Compute Cloud Distance score.

        :param kernel: a kernel
        :type kernel: ee.Kernel
        :param dmax: Maximum distance to calculate the score. If the pixel is
            further than dmax, the score will be 1.
        :type dmax: int
        :param dmin: Minimum distance.
        :type dmin: int
        :param bandmask: the band that contains the mask. Defaults to the first
            band of the image
        :type bandmask: str
        :param bandname: the name of the resulting band. Defaults to
            'cloud_dist'
        :type bandname: str
        :param units: units for the kernel. Can be 'pixels' or 'meters'. Defaults to the latter
        :type units: str
        """
        kernel = kwargs.get('kernel')
        dmax = ee.Number(kwargs.get('dmax'))
        dmin = ee.Number(kwargs.get('dmin'))
        bandmask = kwargs.get('bandmask', 0)
        bandname = kwargs.get('bandname', 'cloud_dist')
        units = kwargs.get('units', 'meters')
        factor = kwargs.get('factor', 0.2)

        if not kernel:
            kernel = ee.Kernel.euclidean(radius=dmax, units=units)
        else:
            kernel = kernel(radius=dmax, units=units)

        cloud_mask = img.mask().select([bandmask])

        # Compute distance to the mask (inverse)
        distance = cloud_mask.Not().distance(kernel)

        # Mask out pixels that are further than d_max
        clip_max_masc = distance.lte(dmax)
        distance = distance.updateMask(clip_max_masc)

        # Mask out initial mask
        distance = distance.updateMask(cloud_mask)

        dmaxi = ee.Image(dmax)
        dmini = ee.Image(dmin)
        factori = ee.Image(factor)

        # Compute score
        pjeDist = ee.Image().expression('1-exp((-dist+dmin)/(dmax*factor))',
                                        {
                                            'dist': distance,
                                            'dmin': dmini,
                                            'dmax': dmaxi,
                                            'factor': factori
                                        }).rename(bandname)

        # Inverse mask to add later
        masc_inv = pjeDist.mask().Not()

        # apply mask2zero (all masked pixels are converted to 0)
        pjeDist = pjeDist.unmask()

        # Add the inverse mask to the distance image
        pjeDist = pjeDist.add(masc_inv)

        # Apply the original mask
        pjeDist = pjeDist.updateMask(cloud_mask)

        return pjeDist.rename(bandname)

    @staticmethod
    def apply(collection, **kwargs):
        return collection.map(lambda img: img.addBands(CloudDist.compute(img, **kwargs)))

    def map(self, collection, **kwargs):
        """ Map function to use in BAP

        :param col: collection
        :type col: geetools.collection.Collection
        """
        col = kwargs.get('col')
        first_band = col.bands[0]
        scale = min([band.scale for band in col.bands])

        maxdist = (scale/2)*510

        # Truncate dmax if goes over the 512 pixels limit
        if self.dmax is None:
            dmax = maxdist
        elif self.dmax > maxdist and self.unit == 'meters':
            dmax = maxdist
        else:
            dmax = self.dmax

        params = dict(
            bandmask = first_band.name,
            kernel = KERNELS_DISTANCE[self.kernel],
            dmin = self.dmin,
            dmax = dmax,
            bandname = self.name,
            units = self.units
        )

        def wrap(img):
            score_img = self.compute(img, **params)

            adjusted_score = self.adjust()(score_img)
            return img.addBands(adjusted_score)

        return collection.map(wrap)


@register(factory)
@register_all(__all__)
class Doy(Score):
    """ Score for the 'Day of the Year (DOY)'

    :param best_doy: the date (MM-DD) that will be prioritized
    :type best_doy: str
    :param season: the season
    :type season: season.Season
    :param name: name for the resulting band
    :type name: str
    """
    def __init__(self, best_doy, season, name="score-best_doy",
                 function='linear', stretch=1, **kwargs):
        super(Doy, self).__init__(**kwargs)
        self.name = name
        self.function = function
        self.stretch = stretch
        self.best_doy = best_doy
        self.season = season

    def adjust(self):
        return lambda img: img

    @staticmethod
    def apply(collection, **kwargs):
        """ Apply best_doy score to every image in a collection.

        :param best_doy: day of year
        :type best_doy: ee.Date
        :param function: the function to use. Can be one of
            'linear' or 'gauss'
        :type function: str
        :return: the parsed collection with a new property called by parameter
            `name` (defaults to 'best_doy').
        :rtype: ee.ImageCollection
        """
        doy = kwargs.get('best_doy')  # ee.Date
        function = kwargs.get('function', 'linear')
        name = kwargs.get('name', 'doy_score')
        stretch = kwargs.get('stretch', 1)
        output_min = kwargs.get('output_min', 0)
        output_max = kwargs.get('output_max', 1)

        # temporary distance property name
        uid = uuid4()
        distance_name = 'distance_{}'.format(uid)

        # function to compute distance
        def distance(img):
            idate = ee.Date(img.date())
            dist = idate.difference(doy, 'day')
            return img.set(distance_name, dist)

        # compute distance to best_doy
        collection = collection.map(distance)

        if function == 'linear':
            result = tools.imagecollection.linearFunctionProperty(
                collection,
                distance_name,
                mean=0,
                output_min= output_min,
                output_max= output_max,
                name=name)

        if function == 'gauss':
            result = tools.imagecollection.gaussFunctionProperty(
                collection,
                distance_name,
                mean=0,
                output_min= output_min,
                output_max= output_max,
                stretch=stretch,
                name=name
            )

        def addBand(img):
            doyn = ee.Number(img.get(name))
            doyband = ee.Image.constant(doyn).rename(name).toFloat()
            return img.addBands(doyband)

        return result.map(addBand)

    def map(self, collection, **kwargs):
        """ Map function to use in BAP

        :param year: the analysing year. Must match the year of the bap
        :type year: int
        """
        range_out = self.range_out
        year = kwargs.get('year')
        date_range = self.season.add_year(year)
        doy = ee.Date(season_module.SeasonDate(self.best_doy).add_year(year))
        doy2 = ee.Date(season_module.SeasonDate(self.best_doy).add_year(year-1))
        condition = date_range.contains(doy)
        best = ee.Number(ee.Algorithms.If(condition, doy, doy2))

        return self.apply(collection, best_doy=best, name=self.name,
                          output_min=range_out[0], output_max=range_out[1],
                          function=self.function, stretch=self.stretch)


@register(factory)
@register_all(__all__)
class AtmosOpacity(Score):
    """ Score for 'Atmospheric Opacity'

    :param range_in: Range of variation for the atmos opacity
    :type range_in: tuple
    :param formula: Distribution formula
    :type formula: Expression
    """
    def __init__(self, range_in=(100, 300), formula=Expression.Exponential,
                 name="score-atm-op", **kwargs):
        super(AtmosOpacity, self).__init__(**kwargs)
        self.range_in = range_in
        self.formula = formula
        self.name = name

    @property
    def expr(self):
        expresion = self.formula(rango=self.range_in)
        return expresion

    def map(self, collection, **kwargs):
        """ Map the score over a collection

        :param col: collection
        :type col: satcol.Collection
        """
        col = kwargs.get('col')
        band = col.getBand('atmos_opacity', 'name')
        if band:
            f = self.expr.map(name=self.name,
                              band=band.name,
                              map=self.adjust(),
                              **kwargs)
        else:
            f = self.empty

        return collection.map(f)


@register(factory)
@register_all(__all__)
class MaskPercent(Score):
    """ This score represents the 'masked pixels cover' for a given area.
    It uses a ee.Reducer so it can consume much EE capacity

    :param band: band of the image that holds the masked pixels
    :type band: str
    :param maxPixels: same param of ee.Reducer
    :type maxPixels: int
    :param include_zero: include pixels with zero value as mask
    :type include_zero: bool
    """
    @staticmethod
    def compute(image, **kwargs):
        """ Core function for Mask Percent Score. Has no dependencies in geebap
        module

        :param image: ee.Image holding the mask
        :type image: ee.Image
        :param geometry: the score will be computed inside this geometry
        :type geometry: ee.Geometry or ee.Feature
        :param scale: the scale of the mask
        :type scale: int
        :param band_name: the name of the resulting band
        :type band_name: str
        :return: An image with one band that holds the percentage of pixels
            with value 0 (not 1) over the total pixels inside the geometry, and
            a property with the same name as the assigned for the band with the
            same percentage
        :rtype: ee.Image
        """
        geometry = kwargs.get('geometry')
        scale = kwargs.get('scale', 1000)
        band_name = kwargs.get('band_name', 'score-maskper')
        max_pixels = kwargs.get('max_pixels', 1e13)
        count_zeros = kwargs.get('count_zeros', False)

        # get band name
        band = ee.String(image.bandNames().get(0))

        # get projection
        projection = image.select([band]).projection()

        # Make an image with all ones
        ones_i = ee.Image.constant(1).reproject(projection).rename(band)

        # manage geometry types
        if isinstance(geometry, (ee.Feature, ee.FeatureCollection)):
            geometry = geometry.geometry()

        # Get total number of pixels
        ones = ones_i.reduceRegion(
            reducer= ee.Reducer.count(),
            geometry= geometry,
            scale= scale,
            maxPixels= max_pixels).get(band)
        ones = ee.Number(ones)

        # select first band, unmask and get the inverse
        mask_image = image.select([band])
        if count_zeros:
            zeros = mask_image.eq(0)
            mask_image = mask_image.updateMask(zeros.Not())

        mask = mask_image.mask()
        mask_not = mask.Not()
        image_to_compute = mask.updateMask(mask_not)

        # Get number of zeros in the given mask_image
        zeros_in_mask =  image_to_compute.reduceRegion(
            reducer= ee.Reducer.count(),
            geometry= geometry,
            scale= scale,
            maxPixels= max_pixels).get(band)
        zeros_in_mask = ee.Number(zeros_in_mask)

        percentage = tools.number.trimDecimals(zeros_in_mask.divide(ones), 4)

        # Make score inverse to percentage
        score = ee.Number(1).subtract(percentage)

        percent_image = ee.Image.constant(score) \
                          .select([0], [band_name])\
                          .set(band_name, score).toFloat()

        return percent_image.clip(geometry)

    def __init__(self, band=None, name="score-maskper", maxPixels=1e13,
                 count_zeros=False, **kwargs):
        super(MaskPercent, self).__init__(**kwargs)
        self.band = band
        self.maxPixels = maxPixels
        self.name = name
        self.count_zeros = count_zeros
        self.sleep = kwargs.get("sleep", 30)

    def map(self, collection, **kwargs):
        """ Map the score over a collection

        :param col: collection
        :type col: satcol.Collection
        :param geom: boundaries geometry
        :type geom: ee.Geometry or ee.Feature
        """
        col = kwargs.get('col')
        geom = kwargs.get('geom')
        minscale = min([band.scale for band in col.bands])
        def wrap(img):
            score = self.compute(img, geometry=geom, scale=minscale,
                                 count_zeros=self.count_zeros)
            prop = score.get(self.name)
            return img.addBands(score).set(self.name, prop)

        return collection.map(wrap)


class MaskPercentKernel(Score):
    """ Mask percent score using a kernel """
    def __init__(self, kernel='square', distance=255, units='pixels',
                 name="score-maskper-kernel", **kwargs):
        """ Initialize score with kernel, distance and units """
        super(MaskPercentKernel, self).__init__(**kwargs)

        self.kernel = kernel
        self.distance = distance
        if self.distance > 255:
            self.distance = 255

        self.units = units
        self.name = name

    @staticmethod
    def _make_kernel(name, radius, units):
        fkernel = KERNELS_BOOL[name]
        return fkernel(radius=radius, units=units)

    @staticmethod
    def compute(img, **kwargs):
        """ Compute kernel mask percent score. Uses the mask of the first band

        :param kernel: the kernel to use (the class, not instance)
        :type kernel: ee.Kernel
        :param size: the kernel size
        :type size: int
        :param: units: units for the kernel ('pixels' or 'meters')
        :type units: str
        :param name: the name of the resulting band
        :type name: str
        """
        kernel = kwargs.get('kernel', 'square')
        size = kwargs.get('size', 255)
        units = kwargs.get('units', 'pixels')
        count_zeros = kwargs.get('count_zeros', False)
        name = kwargs.get('name', 'mask-kernel')

        if units == 'pixels' and size > 255:
            size = 255

        distance = (size*2+1)**2

        Kernel = MaskPercentKernel._make_kernel(kernel, size, units)

        maskband = img.select([0])

        if count_zeros:
            mask = maskband.eq(0).Not()
            maskband = maskband.updateMask(mask)

        count = maskband.reduceNeighborhood(
            'count', Kernel).rename(maskband.bandNames())

        return count.divide(distance).rename(name)

    def map(self, collection, **kwargs):
        def wrap(img):
            score = self.compute(
                img,
                kernel=self.kernel,
                size=self.distance,
                units=self.units,
                name=self.name
            )
            return img.addBands(score)

        return collection.map(wrap)


@register(factory)
@register_all(__all__)
class Satellite(Score):
    """ Score for the satellite

    :param rate: 'amount' of the score that will be taken each step of the
        available satellite list
    :type rate: float
    """
    def __init__(self, ratio=0.05, name="score-sat", **kwargs):
        super(Satellite, self).__init__(**kwargs)
        self.name = name
        self.ratio = ratio

    @staticmethod
    def compute(image, **kwargs):
        colid = kwargs.get('collection_id') # ej: 'COPERNICUS/S2'
        year = kwargs.get('year')
        rate = kwargs.get('ratio', 0.05)
        name = kwargs.get('name', 'sat-score')

        year_str = ee.Number(year).format()
        # List of satellite priority according to year
        prior_list = ee.List(priority.SeasonPriority.ee_relation.get(year_str))

        # Get index of current satellite into the list
        index = prior_list.indexOf(colid)

        # catch if item not in list
        exists = ee.Number(index).neq(-1)

        # 1 - (0.05 * index)
        # EJ: [1, 0.95, 0.9]
        factor = ee.Number(rate).multiply(index)
        factor = ee.Number(ee.Algorithms.If(exists, factor, 1))
        sat_score = ee.Number(1).subtract(factor)

        # Create the score band
        score_img = ee.Image.constant(sat_score).rename(name).toFloat()

        # Write score as an Image Property
        score_img = score_img.set(name.upper(), sat_score)

        return score_img

    def map(self, collection, **kwargs):
        """
        :param col: Collection
        :type col: satcol.Collection
        """
        col = kwargs.get('col')
        year = kwargs.get('year')

        def wrap(img):
            y = ee.Number(year) if year else img.date().get('year')
            score = self.compute(img, collection_id=col.id, year=y,
                                 rate=self.ratio, name=self.name)
            return img.addBands(score).set(self.name, score.get(self.name))

        return collection.map(wrap)


@register(factory)
@register_all(__all__)
class Outliers(Score):
    """ Score for outliers

    Compute a pixel based score regarding to its 'outlier' condition. It
    can use more than one band.

    To see an example, run `test_outliers.py`

    :param bands: name of the bands to compute the outlier score
    :type bands: tuple
    :param process: Statistic to detect the outlier
    :type process: str
    :param dist: 'distance' to be considered outlier. If the chosen process
        is 'mean' the distance is in 'standar deviation' else if it is
        'median' the distance is in 'percentage/100'. Example:

        dist=1 -> min=0, max=100

        dist=0.5 -> min=25, max=75

        etc

    :type dist: int
    """
    def __init__(self, bands, process="median", dist=0.7, name="score-outlier",
                 **kwargs):
        super(Outliers, self).__init__(**kwargs)

        # TODO: param bands is related to the collection used
        self.bands = bands
        self.process = process

        # TODO: distribution
        self.dist = dist
        self.range_in = (0, 1)
        self.name = name
        self.sleep = kwargs.get("sleep", 10)

        # TODO: create `min` and `max` properties depending on the chosen process

    @property
    def bands_ee(self):
        return ee.List(self.bands)

    @property
    def bandslength(self):
        return float(len(self.bands))

    @property
    def increment(self):
        return float(1 / self.bandslength)

    @staticmethod
    def apply(collection, **kwargs):
        """ Determine if pixels are outliers given a collection and parameters

        :param bands: the bands to use for determination
        :type bands: list
        :param reducer: the reducer to use. Can be 'mean' or 'median'
        :type reducer: str
        :param amount: how many stdDev (mean) or percentage (median) to
            determine the upper and lower limit
        :type amount: float
        """
        bands = kwargs.get('bands')
        reducer = kwargs.get('reducer')
        amount = kwargs.get('amount')

        if reducer is None:
            reducer = 'mean'

        if amount is None:
            if reducer == 'mean':
                amount = 0.7
            elif reducer == 'median':
                amount = 0.5

        # MASK PIXELS = 0 OUT OF EACH IMAGE OF THE COLLECTION
        col = collection.map(lambda img: img.selfMask())

        if bands is None:
            bands = ee.Image(col.first()).bandNames()

        # Collection with selected bands only
        selected = col.select(bands)

        # get statistics for mean process
        if reducer == "mean":
            mean = ee.Image(selected.mean())
            std = ee.Image(selected.reduce(ee.Reducer.stdDev()))
            distance = std.multiply(amount)

            mmin = mean.subtract(distance).rename(bands)
            mmax = mean.add(distance).rename(bands)

        # get statistics for median process
        elif reducer == "median":
            mmin = ee.Image(selected.reduce(ee.Reducer.percentile([50-(50*amount)])))
            mmin = mmin.rename(bands)
            mmax = ee.Image(selected.reduce(ee.Reducer.percentile([50+(50*amount)])))
            mmax = mmax.rename(bands)

        def wrap(img):
            # select bands
            bands_i = img.select(bands)

            # condition inside
            condition = bands_i.gte(mmin) \
                .And(bands_i.lte(mmax))

            if reducer == 'median':
                # condition = condition.select(condition.bandNames(), img.bandNames())
                condition = condition.rename(bands)

            condition = condition.Not()
            pout = tools.image.addSuffix(condition, '_outlier')

            return img.addBands(pout) \
                .set('SCORE_OUTLIER_REDUCER', reducer) \
                .set('SCORE_OUTLIER_AMOUNT', amount)

        return collection.map(wrap)


    def map(self, collection, **kwargs):
        """
        :return:
        :rtype: ee.Image
        """
        name = self.name
        increment = self.increment
        reducer = self.process
        amount = self.dist

        outliers = self.apply(collection, bands=self.bands, reducer=reducer,
                              amount=amount)

        pattern = self.bands_ee.map(
            lambda name: ee.String(name).cat('_outlier'))

        def wrap(img):
            out = img.select(pattern)
            # As apply method returns 1 for outliers and the score should be 0
            # for it, turn it upside down
            out = out.Not()
            suma = tools.image.sumBands(out, name)
            final = suma.select(name) \
                .multiply(ee.Image(increment)).rename(name)

            no_out = tools.image.removeBands(img, pattern)

            return no_out.addBands(final)

        return outliers.map(wrap)


@register(factory)
@register_all(__all__)
class Index(Score):
    """ Score for a vegetation index. As higher the index value, higher the
    score.

    :param index: name of the vegetation index. Can be 'ndvi', 'evi' or 'nbr'
    :type index: str
    """
    def __init__(self, index="ndvi", target=0.8, name="score-index",
                 function='linear', stretch=1, **kwargs):
        super(Index, self).__init__(**kwargs)
        self.index = index
        self.range_in = kwargs.get("range_in", (0, 1))
        self.name = name
        self.function = function
        self.target = target
        self.stretch = stretch

    def adjust(self):
        return lambda img: img

    @staticmethod
    def compute(image, **kwargs):
        """ Compute Index Score. Parameters:

        - index (str): the name of the index
        - target (float): the value of the index that will be the output
          maximum
        - range_min (float): the minmum value of the index range
        - range_max (float): the maximum value of the index range
        - out_min (float): the minmum value of the expected output. Defaults to
          zero
        - out_max (float): the maximum value of the expected output. Defaults
          to one
        - function (str): can be 'linear' or 'gauss'. Defaults to 'linear'
        - stretch (float): a stretching value for the normal distribution
        """
        index = kwargs.get('index')
        target = kwargs.get('target')
        range_min = kwargs.get('range_min')
        range_max = kwargs.get('range_max')
        out_min = kwargs.get('out_min', 0)
        out_max = kwargs.get('out_max', 1)
        function = kwargs.get('function', 'linear')
        stretch = kwargs.get('stretch', 1)
        name = kwargs.get('name')

        if function == 'linear':
            result = tools.image.linearFunction(image, index,
                                                 range_min=range_min,
                                                 range_max=range_max,
                                                 mean=target,
                                                 output_min=out_min,
                                                 output_max=out_max,)

        elif function == 'gauss':
            result = tools.image.gaussFunction(image, index,
                                                range_min=range_min,
                                                range_max=range_max,
                                                mean=target,
                                                output_min=out_min,
                                                output_max=out_max,
                                                stretch=stretch)
        else:
            raise ValueError('function parameter must be "linear" or "gauss"')

        return result.rename(name)

    def map(self, collection, **kwargs):
        def wrap(img):
            result = self.compute(img, index=self.index,
                                  function=self.function,
                                  name=self.name,
                                  range_min=self.range_in[0],
                                  range_max=self.range_in[1],
                                  target=self.target,
                                  out_min=self.range_out[0],
                                  out_max=self.range_out[1],
                                  stretch=self.stretch
                                  )
            return img.addBands(result)

        return collection.map(wrap)


@register(factory)
@register_all(__all__)
class MultiYear(Score):
    """ Score for a multiyear (multiseason) composite. Suppose you create a
    single composite for 2002 but want to use images from 2001 and 2003. To do
    that you have to indicate in the creation of the Bap object, but you want
    to prioritize the central year (2002). To do that you have to include this
    score in the score's list.

    :param main_year: central year
    :type main_year: int
    :param season: main season
    :type season: season.Season
    :param ratio: how much score will be taken each year. In the example would
        be 0.95 for 2001, 1 for 2002 and 0.95 for 2003
    :type ration: float
    """

    def __init__(self, season, ratio=0.05, main_year=None, function='linear',
                 stretch=1, name="score-multi", **kwargs):
        super(MultiYear, self).__init__(**kwargs)
        self.main_year = main_year
        self.season = season
        self.ratio = ratio
        self.function = function
        self.name = name
        self.stretch = stretch
        self.year_property = kwargs.get('year_property', 'YEAR_BAP')

    def adjust(self):
        """ redefine adjust method for NOT adjusting """
        return lambda img: img

    @staticmethod
    def apply(collection, **kwargs):
        """ Apply multi year score to every image in a collection.

        :param target_year: target year
        :type target_year: int
        :param function: the function to use. Can be one of
            'linear' or 'gauss'
        :type function: str
        :param name: name for the resulting band
        :type name: str
        :param stretch: stretch parameter for the Gauss function
        :type stretch: float
        :param output_min: minimum desired value for the output
        :type output_min: float
        :param output_max: maximum desired value for the output
        :type output_max: float
        :param year_property: the name of the property that holds the date. If
            None, it will compute the year taken from the image date
        :param year_property: str
        :return: the parsed collection with a new property called by parameter
            `name` (defaults to 'year_score').
        :rtype: ee.ImageCollection
        """
        year = ee.Number(kwargs.get('target_year'))
        function = kwargs.get('function', 'linear')
        name = kwargs.get('name', 'year_score')
        stretch = kwargs.get('stretch', 1)
        output_min = kwargs.get('output_min', 0)
        output_max = kwargs.get('output_max', 1)
        year_property = kwargs.get('year_property')

        # temporary distance property name
        uid = uuid4()
        distance_name = 'distance_{}'.format(uid)

        # function to compute distance
        def distance(img):
            if year_property:
                iyear = ee.Number(img.get(year_property))
            else:
                iyear = ee.Date(img.date()).get('year')

            dist = year.subtract(ee.Number(iyear))
            return img.set(distance_name, dist)

        # compute distance to best_doy
        collection = collection.map(distance)

        if function == 'linear':
            result = tools.imagecollection.linearFunctionProperty(
                collection,
                distance_name,
                mean=0,
                output_min= output_min,
                output_max= output_max,
                name=name)

        elif function == 'gauss':
            result = tools.imagecollection.gaussFunctionProperty(
                collection,
                distance_name,
                mean=0,
                output_min= output_min,
                output_max= output_max,
                stretch=stretch,
                name=name
            )

        else:
            raise ValueError("function must be 'linear' or 'gauss'")

        def addBand(img):
            score = ee.Number(img.get(name))
            scoreband = ee.Image.constant(score).rename(name).toFloat()
            return img.addBands(scoreband)

        final = result.map(addBand)
        return final

    def map(self, collection, **kwargs):
        """ This method keeps only the images included in the seasons

        :param years: the list of all years
        :type years: list
        """
        year = self.main_year or kwargs.get('year')
        range_out = self.range_out

        # Use YEAR_BAP property that is set in the BAP process
        return self.apply(collection, target_year=year, name=self.name,
                          output_min=range_out[0], output_max=range_out[1],
                          function=self.function, stretch=self.stretch,
                          year_property=self.year_property)


@register(factory)
@register_all(__all__)
class Threshold(Score):
    def __init__(self, bands=None, name='score-thres',
                 **kwargs):
        """ Threshold score """
        super(Threshold, self).__init__(**kwargs)

        self.bands = bands
        self.name = name

    @staticmethod
    def compute(img, **kwargs):
        """ Compute the threshold score

        :param thresholds: a dictionary of threshold values for each band. The
            keys of the dict must be the name of the bands, and the value for
            each band MUST be a dict with 2 keys: `min` and `max`. For example:

            ``` python
            threshold = {'B1': {'min': 1000, 'max': 3000}}
            ```
        :type thresholds: dict
        :param name: the name of the resulting band
        :type name: str
        :rtype: ee.Image
        """
        thresholds = kwargs.get('thresholds')
        name = kwargs.get('name', 'score-threshold')

        # Cast
        thresholds = ee.Dictionary(thresholds)

        bands_ee = ee.List(thresholds.keys())
        relation_ee = ee.Dictionary(thresholds)
        length = ee.Number(bands_ee.size())

        def compute_score(band, first):
            score_complete = ee.Image(first)
            img_band = img.select([band])
            min = ee.Dictionary(relation_ee.get(band)).get('min')  # could be None
            max = ee.Dictionary(relation_ee.get(band)).get('max')  # could be None

            score_min = img_band.gte(ee.Image.constant(min))
            score_max = img_band.lte(ee.Image.constant(max))

            final_score = score_min.And(score_max)  # Image with one band

            return tools.image.replace(score_complete, band, final_score)

        scores = ee.Image(bands_ee.iterate(compute_score,
                                           tools.image.empty(0, bands_ee)))

        final_score = tools.image.sumBands(scores, name=name) \
            .divide(ee.Image.constant(length))

        return final_score.select(name)

    def map(self, collection, **kwargs):
        """ map the score over a collection

        :param col: the collection
        :type col: satcol.Collection
        """
        print('score not available temporarely')
        return collection


@register(factory)
@register_all(__all__)
class Medoid(Score):
    def __init__(self, bands=None, discard_zeros=True, name='score-medoid',
                 **kwargs):
        super(Medoid, self).__init__(**kwargs)
        self.name = name
        self.bands = bands
        self.discard_zeros = discard_zeros

    @staticmethod
    def apply(collection, **kwargs):
        return composite.medoidScore(collection, **kwargs)

    def map(self, collection, **kwargs):

        return self.apply(collection, bands=self.bands,
                          discard_zeros=self.discard_zeros,
                          bandname=self.name,
                          normalize=self.normalize)


@register(factory)
@register_all(__all__)
class Brightness(Score):
    def __init__(self, target=1, bands=None, name='score-brightness',
                 function='linear', **kwargs):
        """ Brightness score

        :param target: a percentage target. For exampĺe, target=0.8 means that
            the top score will be for the 80 percent of the maximum brightness
        :type target: float
        """
        super(Brightness, self).__init__(**kwargs)
        if not bands:
            bands = ['green', 'blue', 'red', 'nir', 'swir']
        self.bands = bands
        self.name = name
        self.function = function
        self.target = target

    def adjust(self):
        return lambda img: img

    @staticmethod
    def compute(image, **kwargs):
        """ Compute a brightness score.

        :param bands: the bands to use for brightness
        :type bands: list
        :param target: the brighness value that will take the top score
        :type target: float
        :param function: the function that will be used for computing brighness
            score. Can be 'linear' or 'gauss'
        :type function: str
        :param min_value: minimum value that takes each band
        :type min_value: float
        :param min_value: maximum value that takes each band
        :type min_value: float

        """
        bands = kwargs.get('bands')
        min_value = kwargs.get('min_value')
        max_value = kwargs.get('max_value')
        max_brighness = kwargs.get('max_brightness')
        min_brighness = kwargs.get('min_brightness')
        function = 'linear'
        target = kwargs.get('target')
        output_min = kwargs.get('output_min', 0)
        output_max = kwargs.get('output_max', 1)
        name = kwargs.get('name')
        stretch = kwargs.get('stretch')

        if not bands:
            bands = image.bandNames()

        bands = ee.List(bands)
        length = ee.Number(bands.size())
        img = image.select(bands)

        if min_brighness is not None:
            allmin = ee.Number(min_brighness)
        elif min_value is not None and min_brighness is None:
            min_value = ee.Number(min_value)
            allmin = min_value.multiply(length)
        else:
            msg = "You must specify a minimum value for brightness, either" \
                  " using `min_value` parameter or `min_brightness` parameter"
            raise ValueError(msg)

        if max_brighness:
            allmax = ee.Number(max_brighness)
        elif max_value and not max_brighness:
            max_value = ee.Number(max_value)
            allmax = max_value.multiply(length)
        else:
            msg = "You must specify a maximum value for brightness, either" \
                  " using `max_value` parameter or `max_brightness` parameter"
            raise ValueError(msg)

        brightness = img.reduce('sum')

        if function == 'linear':
            result = tools.image.linearFunction(
                image=brightness,
                band='sum',
                range_min=allmin,
                range_max=allmax,
                mean=target,
                name=name,
                output_min=output_min,
                output_max=output_max
            )
        elif function == 'gauss':
            result = tools.image.gaussFunction(
                image=brightness,
                band='sum',
                range_min=allmin,
                range_max=allmax,
                mean=target,
                name=name,
                output_min=output_min,
                output_max=output_max,
                stretch=stretch
            )

        return result

    def map(self, collection, **kwargs):
        """ Map score over a collection.

        :param col: collection
        :type col: satcol.Collection
        """
        col = kwargs.get('col')

        mx = max([b.max for b in col.bands if b.name in self.bands])

        length = len(self.bands)

        target = mx * length * self.target

        def wrap(img):
            score = self.compute(
                image=img,
                target=target,
                bands=self.bands,
                max_value=mx,
                min_value=0,
                function=self.function,
                name=self.name,
                output_min=self.range_out[0],
                output_max=self.range_out[1],
            )

            return img.addBands(score)

        newcol = collection.map(wrap)
        return newcol


# -*- coding: utf-8 -*-
"""
Package to generate a "Best Available Pixel" image in Google Earth Engine
"""

from __future__ import absolute_import, division, print_function
from ._version import __version__

__all__ = (
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
)

__title__ = "BestAvailablePixel"
__summary__ = "Generate a 'Best Available Pixel' image in Google Earth Engine"
__uri__ = "https://github.com/fitoprincipe/geebap"
__author__ = "Rodrigo E. Principe"
__email__ = "rprincipe@ciefap.org.ar"

__license__ = "GNU GENERAL PUBLIC LICENSE, Version 3"
__copyright__ = "Rodrigo E. Principe"

try:
    from . import bap, date, expgen, expressions, filters, functions,\
        ipytools, masks, regdec, scores, season, sites

    from .bap import Bap
    from .priority import SeasonPriority
    from .season import Season
except ImportError:
    pass


# -*- coding: utf-8 -*-
import ee
from geetools import collection


def get_id_col(id):
    """ get the corresponding collection of the given id """
    try:
        col = collection.IDS[id]
    except IndexError:
        raise IndexError
    else:
        return col


def get_col_id(col):
    return collection.IDS.index(col.id)


def get_col_id_image(col, name='col_id'):
    return ee.Image.constant(get_col_id(col)).rename(name).toUint8()


def drange(ini, end, step=1, places=0):
    """ Create a range of floats

    :param ini: initial value (as in range)
    :param end: final value (as in range)
    :param step: Similar to range, except that if decimal places are specified
        the step is done between decimal places.
    :param places: decimal places
    :return: range
    :rtype: list
    """
    factor = 10**places if places>0 else 1
    ini *= factor
    end = end * factor - factor + 1
    result = [float(val) / factor for val in range(int(ini), int(end), step)]
    return result


def replace_duplicate(list, separator="_"):
    """ replace duplicated values from a list adding a suffix with a number

    :param list: list to be processed
    :type list: list
    :param separator: string to separate the name and the suffix
    :type separator: str
    :return: new list with renamed values
    :rtype: list
    """
    def wrap(a):
        newlist = [a[0]]  # list with first element
        for i, v in enumerate(a):
            if i == 0: continue  # skip first element
            # t = a[i+1:]
            if v in newlist:
                if separator in v:
                    two = v.split(separator)
                    orig = two[0]
                    n = int(two[1])
                    newval = "{}{}{}".format(orig, separator, n+1)
                else:
                    newval = v+separator+"1"
                newlist.append(newval)
            else:
                newlist.append(v)
        return newlist

    new = wrap(list)
    while(new != list):
        list = new[:]
        new = wrap(new)
    return(new)


def nirXred(nir="NIR", red="RED", output="nirXred"):
    """ Creates a NIR times RED band

    :param nir: name of the NIR band
    :type nir: str
    :param red: name of the RED band
    :type red: str
    :param output: name of the output band
    :type output: str
    :return:
    :rtype: function
    """
    def wrap(img):
        n = img.select([nir])
        r = img.select([red])
        nirXred = n.multiply(r).select([0], [output])
        return img.addBands(nirXred)
    return wrap


def unmask_slc_off(image):
    """ Unmask pixels that are masked in ALL bands """
    mask = image.mask()
    reduced = mask.reduce('sum')
    slc_off = reduced.eq(0)
    unmasked = image.unmask()
    newmask = mask.where(slc_off, 1)
    return unmasked.updateMask(newmask)


# -*- coding: utf-8 -*-
import ee
from collections import OrderedDict


def is_leap(year):
    if isinstance(year, (int, float)):
        mod = year%4
        return True if mod == 0 else False
    elif isinstance(year, (ee.Number,)):
        mod = year.mod(4)
        return mod.Not()


def _rel_month_day(leap_year):
    months = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    pairs = zip(months, days)

    month_day = OrderedDict()
    for m, d in pairs:
        if leap_year and m == 2:
            d = 29
        month_day[m] = d
    return month_day


class SeasonDate(object):
    """ A simple class to hold dates as MM-DD """
    def __init__(self, date):
        self.date = date

        # Check if format is valid
        self.check_valid()

    @property
    def month(self):
        return int(self.date.split('-')[0])

    @property
    def day(self):
        return int(self.date.split('-')[1])

    @property
    def day_of_year(self):
        """ Day of the year """
        month = self.month
        day = self.day

        if month == 1:
            return day
        else:
            ini = 31

        for rel_month, days in _rel_month_day(True).items():
            if rel_month == 1: continue

            if rel_month != month:
                ini += days
            else:
                break

        return ini + day

    def add_year(self, year):
        """ Just add the year """
        if not is_leap(year) and self.date == '02-29':
            msg = "Year {} is leap, hence it does't contain day 29 in february"
            raise ValueError(msg.format(year))
        return '{}-{}'.format(year, self.date)

    def check_valid(self):
        """ Verify if the season date has right format """
        if not isinstance(self.date, str):
            mje = "Dates in Season must be strings with format: MM-dd, found " \
                  "{}"
            raise ValueError(mje.format(self.date))

        split = self.date.split("-")
        assert len(split) == 2, \
            "Error in Season {}: month and day must be divided by '-' " \
            "and with the following format --> MM-dd".format(self.date)
        m = int(split[0])
        d = int(split[1])

        if m < 1 or m > 12:
            raise ValueError(
                "Error in Season {}: Month must be greater than 1 and less "
                "than 12".format(self.date))

        maxday = _rel_month_day(True)[m]
        if d < 1 or d > maxday:
            raise ValueError(
                "Error in Season {}: In month {} the day must be less or "
                "equal than {}".format(self.date, m, maxday))

        return True


class Season(object):
    """ Growing season """
    def __init__(self, start, end):
        self.start = start
        self.end = end

    # START
    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        if value is not None:
            if not isinstance(value, SeasonDate):
                try:
                    value = SeasonDate(value)
                except:
                    raise ValueError('start date must be an instance of SeasonDate')
            self._start = value
        else:
            raise ValueError('initial date is required')

    # END
    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        if value is not None:
            if not isinstance(value, SeasonDate):
                try:
                    value = SeasonDate(value)
                except:
                    raise ValueError('end date must be an instance of SeasonDate')

            self._end = value
        else:
            raise ValueError('end date is required')

    @property
    def over_end(self):
        """ True if the season goes over the end of the year """
        if self.start.day_of_year >= self.end.day_of_year:
            return True
        else:
            return False

    @property
    def range_in_days(self):
        return abs(self.start.difference(self.end, self.over_end))

    def add_year(self, year):
        year = ee.Number(year)
        if self.over_end:
            start_year = year.subtract(1)
        else:
            start_year = ee.Number(year)
        end_year = ee.Number(year)

        sday = self.start.day
        eday = self.end.day

        # look for feb 29h in non leap
        if not is_leap(year):
            if self.start.month == 2 and sday == 29:
                sday = 28
            if self.end.month == 2 and eday == 29:
                eday = 28

        start = ee.Date.fromYMD(start_year, self.start.month, sday)
        end = ee.Date.fromYMD(end_year, self.end.month, eday)
        daterange = ee.DateRange(ee.Date(start), ee.Date(end))
        return daterange


# -*- coding: utf-8 -*-
""" Module holding custom filters for image collections """
from abc import ABCMeta, abstractmethod
from .regdec import *

__all__ = []
factory = {}


class Filter(object):
    """ Abstract Base class for filters """
    __metaclass__ = ABCMeta
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def apply(self, colEE, **kwargs):
        pass

@register(factory)
@register_all(__all__)
class CloudCover(Filter):
    """ Cloud cover percentage filter

    :param percent: all values over this will be filtered. Goes from 0
        to 100
    :type percent: int
    :param kwargs:
    """
    def __init__(self, percent=70, **kwargs):
        super(CloudCover, self).__init__(**kwargs)
        self.percent = percent
        self.name = 'CloudCover'

    def apply(self, collection, **kwargs):
        """ Apply the filter

        :param colEE: the image collection to apply the filter
        :type colEE: ee.ImageCollection
        :param col: if the collection is given, it will take the name of the
            attribute holding the 'cloud cover' information
        :type col: satcol.Collection
        :param prop: if the collection is not given, the name of the attribute
            holding the 'cloud cover' information can be given
        :type prop: str
        :param kwargs:
        :return:
        """
        col = kwargs.get("col")
        if col.cloud_cover:
            return collection.filterMetadata(col.cloud_cover, "less_than",
                                             self.percent)
        elif 'prop' in kwargs.keys():
            prop = kwargs.get('prop')
            return collection.filterMetadata(prop, 'less_than', self.percent)
        else:
            return collection


@register(factory)
@register_all(__all__)
class MaskCover(Filter):
    """ This mask can only be used AFTER computing mask percentage score
    (`scores.MaskPercent`). This score writes an attribute to the image
    which contains 'mask percentage' of the given area (see score's docs).
    This filter uses this value to apply a filter. I's similar to common
    'cloud cover' filter, but takes account only the given area and not
    the whole scene.

    :param percent: all values over this will be filtered. Goes from 0 to 1
    :type percent: float
    :param prop: name of the property containing the 'mask percentage'
    :param prop: str
    :param kwargs:
    """
    def __init__(self, percent=0.7, prop="score-maskper", **kwargs):
        super(MaskCover, self).__init__(**kwargs)
        self.percent = percent
        self.prop = prop
        self.name = 'MaskCover'

    def apply(self, collection, **kwargs):
        """ Apply the filter

        :param collection: the image collection to apply the filter
        :type collection: ee.ImageCollection
        :return:
        """
        return collection.filterMetadata(self.prop, "less_than", self.percent)



# -*- coding: utf-8 -*-
""" Util functions """


def get_init_params(obj):
    init_params = obj.__init__.__code__.co_varnames
    obj_params = obj.__dict__.items()
    return {param: value for param, value in obj_params if param in init_params}


def object_init(obj):
    init_params = get_init_params(obj)
    return '{}(**{})'.format(obj.__class__.__name__, init_params)


def serialize(obj, name=None, result=None):
    """ Serialize an object to a dict """
    if result is None:
        result = {}

    def make_name(obj, name=None):
        objname = obj.__class__.__name__
        if name is None:
            # name = objname
            return '({})'.format(objname)
        return '{} ({})'.format(name, objname)

    name = make_name(obj, name)

    try:
        # If it is an object, it has a __dict__
        obj_attr = obj.__dict__
    except AttributeError:
        # If it's NOT an object
        if isinstance(obj, (tuple, list)):
            newlist = []
            for element in obj:
                newlist.append(serialize(element))
            result[name] = newlist
        elif isinstance(obj, (dict,)):
            newdict = {}
            for key, element in obj.items():
                newdict[key] = serialize(element)
            result[name] = newdict
        else:
            result[name] = obj
    else:
        # If it IS an object
        attrs = {}
        result[name] = attrs
        for attr_name, attr_value in obj_attr.items():
            serialize(attr_value, attr_name, attrs)

    return result



# -*- coding: utf-8 -*-
""" Common masks to use in BAP process """
from geetools import cloud_mask

def _get_function(col, band, option, renamed=False):
    """ Get mask function for given band and option """
    band_options = col.bitOptions(renamed)
    f = lambda img: img
    if band in band_options:
        bit_options = band_options[band]
        if option in bit_options:
            f = lambda img: col.applyMask(img, band, [option],
                                           renamed=renamed)
    return f


class Mask(object):
    """ Compute common masks regarding the given collection. Looks for
    pixel_qa, BQA, sr_cloud_qa and QA60 bands in that order """
    def __init__(self, options=('cloud', 'shadow', 'snow')):
        self.options = options
        self.bands = ['pixel_qa', 'BQA', 'sr_cloud_qa', 'QA60']

    def map(self, collection, **kwargs):
        """ Map the mask function over a collection

        :param collection: the ImageCollection
        :type collection: ee.ImageCollection
        :param renamed: whether the collection is renamed or not
        :type renamed: bool
        :param col: the EE Collection
        :type col: geetools.collection.Collection
        :return: the ImageCollection with all images masked
        :rtype: ee.ImageCollection
        """
        col = kwargs.get('col')
        renamed = kwargs.get('renamed', False)
        for opt in self.options:
            for band in self.bands:
                f = _get_function(col, band, opt, renamed)
                collection = collection.map(f)

        return collection


class Hollstein(object):
    """ Compute Hollstein mask for Sentinel 2 """
    def __init__(self, options=('cloud', 'shadow', 'snow')):
        self.options = options

    def map(self, collection, **kwargs):
        """ Map the mask function over a collection

        :param collection: the ImageCollection
        :type collection: ee.ImageCollection
        :param renamed: whether the collection is renamed or not
        :type renamed: bool
        :param col: the EE Collection
        :type col: geetools.collection.Collection
        :return: the ImageCollection with all images masked
        :rtype: ee.ImageCollection
        """
        col = kwargs.get('col')
        renamed = kwargs.get('renamed', False)

        bands = []
        for band in ['aerosol', 'blue', 'green', 'red_edge_1', 'red_edge_2',
                     'red_edge_3', 'red_edge_4', 'water_vapor', 'cirrus',
                     'swir']:
            if renamed:
                bands.append(band)
            else:
                bands.append(col.get_band(band, 'name').id)

        if 'hollstein' in col.algorithms:
            f = lambda img: cloud_mask.applyHollstein(img, self.options,
                                                       *bands)
            return collection.map(f)
        else:
            return lambda img: img


# -*- coding: utf-8 -*-
import ee

from . import expgen
from .functions import drange
from geetools import tools
import math
import simpleeval as sval
import numpy as np
import copy

# FUNCIONES PARA simpleeval
CUSTOM_FUNCTIONS = {"sqrt": math.sqrt,
                    "exp": math.exp,
                    "max": max,
                    "min": min}

CUSTOM_NAMES = {"pi": math.pi}

sval.DEFAULT_FUNCTIONS.update(CUSTOM_FUNCTIONS)
sval.DEFAULT_NAMES.update(CUSTOM_NAMES)


class Expression(object):
    # TODO: Limitante: si hay mas de una variable
    """ El método principal de Expresiones() es map(**kwargs), el cual define
    la funcion que se usará en ImageCollection.map()

    :param expresion: Expression en formato texto. Se pueden usar algunas
        variables calculadas de la siguiente forma:

        - {max}: max_result valor del range
        - {min}: minimo valor del range
        - {mean}: mean del range
        - {std}: desvío estandar del range
        - {max_result}: max_result valor de los resultados

        La variable puede ser solo una, que al momento de aplicar el método
        map() se definirá de donde proviene, y debe incluirse en la expresión
        de la siguiente forma: {var}. Ejemplo:

        |
        `expr = "{var}*{mean}/{std}+40"`

        `e = Expression(expression=expr, range=(0, 100))`

        `print e.format_local()`

        >> {var}*50.0/28.8963665536+40

    :type expresion: str
    :param rango: Rango de valores entre los que varía la variable
    :type rango: tuple
    :param normalizar: Hacer que los valores (resultado) oscilen entre
        0 y 1 dividiendo los valores por el max_result. Si se quiere
        normalize se debe proveer un range.
    :type normalizar: bool
    :param kargs: argumentos 'keywords', ej: a=1, b=2

    :Propiedades fijas:
    :param media: Si se provee un range, es la mean aritmetica, sino None
    :type media: float
    :param std: Si se provee un range es el desvío estandar, sino None
    :type std: float
    :param maximo: Determinar el max_result resultado posible. Aplicando la
        expression localmente con la funcion eval()
    :type maximo: float

    :Uso:

    .. code:: python

        # Defino las expresiones con sus rangos
        exp = Expression.Exponential(range=(0, 100))
        exp2 = Expression.Exponential(range=(0, 5000), normalize=True)

        # Defino las funciones, indicando el name de la band resultante
        # y la variable que se usará para el cálculo
        fmap = exp.map("otronombre", prop="CLOUD_COVER")
        fmap2 = exp2.map("name", band="NIR")

        # Mapeo los resultados
        newcol = imgcol.map(fmap)
        newcol2 = imgcol.map(fmap2)

    :Operadores:
    :\+ \- * / % **:    Add, Subtract, Multiply, Divide, Modulus, Exponent
    :== != < > <= >=:   Equal, Not Equal, Less Than, Greater than, etc.

    :Constantes:
    :pi: pi (3.1415)
    :e: euler (2.71)

    :Funciones admitidas:
    :max: max_result valor entre dos argumentos
    :min: minimo valor entre dos argumentos
    :exp: exponencial de e: e^argumento
    :sqrt: raiz cuadrada del argumento
    """
    def __init__(self, expression="{var}", normalize=False, range=None,
                 name=None, **kwargs):
        self.expression = expression
        self.range = range
        self._normalize = normalize
        self.params = kwargs
        self._max = kwargs.get("max")
        self._min = kwargs.get("min")
        self._std = kwargs.get("std")
        self._mean = kwargs.get("mean")
        self.name = name

    def format_local(self):
        """ Reemplaza las variables de la expression por los valores asignados
        al objeto, excepto la variable 'var' que la deja como está porque es
        la que variará en la expression de EE"""
        # print self.expression
        # reemplaza las variables estadisticas
        params = copy.deepcopy(self.params)
        params["max"] = self.max
        params["min"] = self.min
        params["mean"] = self.mean
        params["std"] = self.std

        return self.expression.format(var="{var}", **params)

    def format_ee(self):
        """ Reemplaza las variables de la expression por los valores asignados
        al objeto y genera la expression lista para usar en Earth Engine """
        # reemplaza las variables estadisticas
        params = copy.deepcopy(self.params)
        params["max"] = self.max
        params["min"] = self.min
        params["mean"] = self.mean
        params["std"] = self.std

        expr = self.expression.format(var="'var'", **params)

        return expgen.ExpGen.parse(expr)


    @staticmethod
    def adjust(name, valor):
        """ Ajusta el valor de la band resultante multipliandolo por 'valor'

        :param valor: Valor de adjust
        :type valor: float
        :param name: name de la band que contiene el valor a ajustar
        :type name: str
        :return: La funcion para map()
        :rtype: function
        """
        def wrap(img):
            band = img.select(name).multiply(valor)
            return tools.image.replace(img, name, band)

        return wrap

    @property
    def normalize(self):
        """
        :rtype: bool
        """
        return self._normalize

    @normalize.setter
    def normalize(self, value):
        """ Metodo para setear el valor de la propiedad 'normalize' """
        if type(value) is bool and type(self.range) is tuple:
            self._normalize = value
        else:
            self._normalize = False
            print("If you want to normalize the function, the range must be "
                  " a tuple")

    # ESTADISTICAS DEL RANGO
    @property
    def mean(self):
        if type(self.range) is tuple:
            r = drange(self.range[0], self.range[1] + 1, places=1)
            return np.mean(r)
        elif self._mean:
            return self._mean
        else:
            raise ValueError("To determine the mean the 'range' param must be "
                             "a tuple")

    @property
    def std(self):
        if type(self.range) is tuple:
            r = drange(self.range[0], self.range[1] + 1, places=1)
            return np.std(r)
        elif self._std:
            return self._std
        else:
            raise ValueError("To determine the std the 'range' param must be "
                             "a tuple")

    @property
    def max_result(self):
        """ Determinar el max_result resultado posible. Aplicando la expression
        localmente con la funcion eval()

        :return:
        """
        if type(self.range) is tuple:
            rango = self.range
        elif self._max and self._min:
            rango = (self._min, self._max)
        else:
            raise ValueError("To determine the max result the 'range' param "
                             "must be a tuple")

        r = drange(rango[0], rango[1]+1, places=1)
        lista_result = [self.eval(var) for var in r]
        maximo = max(lista_result)
        return maximo

    @property
    def max(self):
        """ Maximo valor del range """
        val = self.range[1] if self.range else self.params.get("max", None)
        return val

    @property
    def min(self):
        """ Minimo valor del range """
        val = self.range[0] if self.range else self.params.get("min", None)
        return val

    def eval(self, var):
        """ Metodo para aplicar la funcion localmente con un valor dado

        :param var: Valor que se usara como variable
        :return: el resultado de evaluar la expression con un valor dado
        :rtype: float
        """
        expr = self.format_local()
        expr = expr.format(var=var)
        result = sval.simple_eval(expr)
        return result

    def eval_normalized(self, var):
        """ Metodo para aplicar la funcion normalizada (resultado entre 0 y 1)
        localmente con un valor dado. No influye el parametro 'normalize'

        :param var: Valor que se usara como variable
        :return: el resultado de evaluar la expression con un valor dado
        :rtype: float
        """
        e = self.format_local()
        expr = "({e})/{maximo}".format(e=e, maximo=self.max_result)
        expr = expr.format(var=var)
        result = sval.simple_eval(expr)
        return result

    def map(self, name="expression", band=None, prop=None, eval=None,
            map=None, **kwargs):
        """ Funcion para mapear el resultado de la expression

        :param name: name que se le dara a la band de la imagen una vez
            calculada la expression
        :type name: str
        :param band: name de la band que se usara como valor variable
        :type band: str
        :param prop: name de la propiedad que se usara como valor variable
        :type prop: str
        :param eval: funcion para aplicar a la variable. Si la variable es el
            valor de una band, entonces el argumento de la funcion será
            esa band, y si es una propiedad, el argumento será la propiedad.
        :type eval: function
        :param map: funcion para aplicarle al valor final. Puede usarse para
            hacer un adjust o ponderacion. El argumento de la funcion debe ser
            la imagen con la band agregada
        :type map: function
        :return: la funcion para map()
        :rtype: function
        """
        # Define la funcion para aplicarle a la variable
        if eval is None:
            func = lambda x: x
        elif callable(eval):
            func = eval
        else:
            raise ValueError("el parametro 'eval' debe ser una funcion")

        # Define la funcion para aplicar
        # print "map:", map

        if map is None:
            finalf = lambda x: x
        elif callable(map):
            finalf = map
        else:
            raise ValueError("el parametro 'map' debe ser una funcion")

        # reemplazo las variables de la expression
        expr = self.format_ee()

        # Normalizar
        if self.normalize:
            expr = "({e})/{maximo}".format(e=expr, maximo=self.max_result)
        else:
            expr = expr

        # print "name", name
        # print "propiedad", prop
        # print "expression", expr
        # Define la funcion de retorno según si se eligio una propiedad o una band
        if prop is None and band is not None:  # BANDA
            def wrap(img):
                # Selecciono los pixeles con valor distinto de cero
                ceros = img.select([0]).eq(0).Not()

                # aplico la funcion 'eval' a la band
                variable = func(img.select(band))
                # aplico la expression
                '''
                calculo = img.expression(expr,
                                         dict(var=variable, **self.params))
                '''

                calculo = img.expression(expr, dict(var=variable))

                # renombro
                calculo = calculo.select([0], [name])
                # aplico la funcion final sobre la imagen completa
                imgfinal = finalf(img.addBands(calculo))
                # retorno la imagen con la band agregada
                return imgfinal.updateMask(ceros)
        elif band is None and prop is not None:  # PROPIEDAD
            def wrap(img):
                # Selecciono los pixeles con valor distinto de cero
                ceros = img.select([0]).eq(0).Not()
                # aplico la funcion 'eval' a la propiedad
                propval = func(ee.Number(img.get(prop)))
                # aplico la expression
                '''
                calculo = img.expression(expr,
                                         dict(var=propval, **self.params))
                '''
                calculo = img.expression(expr, dict(var=propval))

                # renombro
                calculo = calculo.select([0], [name])
                # aplico la funcion final sobre la imagen completa
                imgfinal = finalf(img.addBands(calculo))
                # imgfinal = img.addBands(calculo)
                # retorno la imagen con la band agregada
                return imgfinal.updateMask(ceros)
        else:
            raise ValueError("la funcion map debe ser llamada con \
                             'band' o 'prop'")

        return wrap

    @classmethod
    def Exponential(cls, a=-10, range=(0, 100), **kwargs):
        """ Funcion Exponential

        :USO:

        :param var: valor variable
        :param media: valor de la mean aritmetica de la variable
        :param a: constante a. El signo determina si el max_result está al final
            de la serie (positivo) o al principio (negativo)
        :param b: constante b. Determina el punto de quiebre de de la curva.
            Cuando es cero, el punto esta en la mean de la serie. Cuando es
            positivo se acerca al principio de la serie, y cuando es negativo
            al final de la serie.
        """
        # DETERMINO LOS PARAMETROS SEGUN EL RANGO DADO SI EXISTIERA
        exp = "1.0-(1.0/(exp(((min({var}, {max})-{mean})*(1/{max}*{a})))+1.0))"
        return cls(expression=exp, a=a, range=range, name="Exponential",
                   **kwargs)

    @classmethod
    def Normal(cls, range=(0, 100), ratio=-0.5, **kwargs):
        """ Campana de Normal

        :param rango: Rango entre los que oscilan los valores de entrada
        :type rango: tuple
        :param ratio: factor de 'agusamiento' de la curva. Debe ser menor a
            cero. Cuanto menor sea, mas 'fina' será la curva
        :type ratio: float
        :param kwargs:
        :return:
        """
        if ratio > 0:
            print("el ratio de la curva gaussiana debe ser menor a cero, "
                  "convirtiendo..")
            ratio *= -1
        if not isinstance(range, tuple):
            raise ValueError("el range debe ser una tupla")

        exp = "exp(((({var}-{mean})/{std})**2)*{ratio})/(sqrt(2*pi)*{std})"
        return cls(expression=exp, range=range, ratio=ratio,
                   name="Normal", **kwargs)


# -*- coding: utf-8 -*-

import ee
ee.Initialize()

from geebap import scores
from geetools import tools, collection

maxDiff = None
band = "B1"

pol = ee.Geometry.Polygon(
    [[[-71.56928658485413, -43.356795174869426],
      [-71.56932950019836, -43.357879484308015],
      [-71.56781673431396, -43.35791848788363],
      [-71.56779527664185, -43.35682637828949]]])

p = ee.Geometry.Point([-71.56871795654297,
                       -43.35720861888331])

col = collection.Landsat8TOA()
colEE = col.collection
colEE = colEE.filterBounds(p)

image = ee.Image(colEE.first()).clip(pol).select([band])

# MASK OUT SOME PIXELS
condition = image.gte(0.13)
image = image.updateMask(condition)

def test_default():
    # SCORE
    score = scores.MaskPercent(band)
    newimg = score.compute(image, geometry=pol, scale=30)

    maskpercent_prop = newimg.get(score.name).getInfo()
    maskpercent_pix = tools.image.getValue(newimg, p, side='client')[score.name]

    assert maskpercent_prop == 0,5625
    assert maskpercent_pix == 0,5625







# -*- coding: utf-8 -*-

import ee
ee.Initialize()
from geebap import scores
from geetools import tools, collection


maxDiff = None
band = "B1"
p = ee.Geometry.Point(-71.5, -42.5)

col = collection.Landsat8TOA().collection
col = col.filterBounds(p).filterDate("2016-11-15", "2017-03-15")

# TEST: MASK VALUES LESS THAN 0.11 ##
def mask(img):
    m = img.select(band).lte(0.11)
    return img.updateMask(m.Not())

col = col.map(mask).map(lambda img: img.unmask())


def test_median():
    '''
    # SECTION TO GENERATE THE RESULT TO COMPARE
    def listval(band):
        def wrap(img, it):
            val = img.reduceRegion(ee.Reducer.first(), p, 30).get(band)
            return ee.List(it).add(ee.Number(val))
        return wrap

    list = col.iterate(listval(band), ee.List([]))

    # Sort list of values
    list = ee.List(list).sort()

    # assertEqual(list.getInfo(), val_list)

    # SCORE
    mean = ee.Number(list.reduce(ee.Reducer.mean()))
    std = ee.Number(list.reduce(ee.Reducer.stdDev()))

    min = mean.subtract(std)
    max = mean.add(std)

    def compute_score(el):
        el = ee.Number(el)
        cond = el.gte(min).And(el.lte(max))

        return ee.Algorithms.If(cond,
        ee.List([ee.Number(el).multiply(10000).format('%.0f'), 1]),
        ee.List([ee.Number(el).multiply(10000).format('%.0f'), 0]))

    to_compare = ee.Dictionary(ee.List(list.map(compute_score)).flatten()).getInfo()
    '''

    # OUTLIER SCORE
    score = scores.Outliers((band,))
    newcol = score.map(col)

    to_compare = {
        "1210": 1, "7579": 0, "1159": 1, "1399": 1, "1171": 1, "1240": 1,
        "1200": 1, "1132": 0, "1328": 1, "1123": 0, "1168": 1, "1150": 1,
        "1178": 1, "0": 0, "1874": 0, "1846": 1, "3090": 0, "1100": 0,
        "1341": 1, "1249": 1, "1866": 1}

    # OUTLIERS SCORE
    val_dict = tools.imagecollection.getValues(newcol, p, 30, side='client')

    # OUTLIER SCORE
    compare = [(str(int(round(val[band]*10000))), val["score-outlier"]) for key, val in val_dict.items()]
    compare = dict(compare)

    assert to_compare == compare


def test_mean():
    score = scores.Outliers((band,), 'mean')
    newcol = score.map(col)

    to_compare = {
        "1210": 1, "7579": 0, "1159": 1, "1399": 1, "1171": 1, "1240": 1,
        "1200": 1, "1132": 1, "1328": 1, "1123": 1, "1168": 1, "1150": 1,
        "1178": 1, "0": 0, "1874": 1, "1846": 1, "3090": 0, "1100": 1,
        "1341": 1, "1249": 1, "1866": 1}

    # OUTLIERS SCORE
    val_dict = tools.imagecollection.getValues(newcol, p, 30, side='client')

    # OUTLIER SCORE
    compare = [(str(int(round(val[band]*10000))), val[score.name]) for key, val in val_dict.items()]
    compare = dict(compare)

    assert to_compare == compare


# -*- coding: utf-8 -*-

import ee
ee.Initialize()
from geebap import scores
from geetools import tools

img = ee.Image('LANDSAT/LC08/C01/T1_SR/LC08_231090_20130414')

p1 = ee.Geometry.Point([-71.51, -43.27])
p2 = ee.Geometry.Point([-71.53, -43.271])
p3 = ee.Geometry.Point([-71.547, -43.274])

region = ee.Geometry.Polygon(
    [[[-71.57, -43.25],
      [-71.57, -43.29],
      [-71.49, -43.29],
      [-71.49, -43.25]]])

original_value1 = 308
original_value2 = 101
original_value3 = 5701


def test_minmax():

    thres = scores.Threshold()
    newimg = thres.compute(img, thresholds={
        'B4': {'min':150, 'max':2000}
    }, name=thres.name)

    val1 = tools.image.getValue(newimg, point=p1, scale=30, side='client')[thres.name]
    val2 = tools.image.getValue(newimg, point=p2, scale=30, side='client')[thres.name]
    val3 = tools.image.getValue(newimg, point=p3, scale=30, side='client')[thres.name]

    assert val1 == 1
    assert val2 == 0
    assert val3 == 0



# -*- coding: utf-8 -*-

import ee
ee.Initialize()
from geebap import scores, bap, season, masks, filters


# FILTERS
filter = filters.CloudCover()

# MASKS
clouds = masks.Mask()

# SEASON
seas = season.Season('11-15', '02-15')

# SCORES
psat = scores.Satellite()
pop = scores.AtmosOpacity()
pmascpor = scores.MaskPercent()
pindice = scores.Index()
pout = scores.Outliers(("ndvi",))
pdoy = scores.Doy('01-15', seas)
thres = scores.Threshold()

# SITES
site = ee.Geometry.Polygon(
    [[[-71.78, -42.79],
      [-71.78, -42.89],
      [-71.57, -42.89],
      [-71.57, -42.79]]])
centroid = site.centroid()

def test_bap2016_0():
    pmulti = scores.MultiYear(2016, seas)
    objbap = bap.Bap(season=seas,
                     scores=(pindice, pmascpor, psat,
                             pout, pop, pdoy, thres,
                             pmulti),
                     masks=(clouds,),
                     filters=(filter,),
                     )

    composite = objbap.build_composite_best(2016, site, indices=("ndvi",))

    assert isinstance(composite, ee.Image) == True



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

here = os.path.dirname(os.path.abspath(__file__))
version_ns = {}
with open(os.path.join(here, 'geebap', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

# the setup
setup(
    name='geebap',
    version=version_ns['__version__'],
    description='Generate a "Best Available Pixel (BAP)" composite in Google '\
                'Earth Engine (GEE)',
    # long_description=read('README'),
    url='',
    author='Rodrigo E. Principe',
    author_email='rprincipe@ciefap.org.ar',
    license='GNU',
    keywords='google earth engine raster image processing gis satelite',
    packages=find_packages(exclude=('docs', 'bap_env')),
    include_package_data=True,
    install_requires=['requests',
                      'simpleeval',
                      'numpy',
                      'geetools>=0.2.2'],
    extras_require={
    'dev': [],
    'docs': [],
    'testing': [],
    },
    classifiers=['Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'License :: OSI Approved :: MIT License',],
    )



# -*- coding: utf-8 -*-
""" Generate expression compatible with Google Earth Engine """
import simpleeval as sval
import ast
import math


class ExpGen(object):
    def __init__(self):
        pass

    @staticmethod
    def max(a, b):
        """ Generates the expression max(a, b) to insert on GEE expression

        :param a: one value to compare
        :param b: the other value to compare
        :return: expression
        :rtype: str
        """

        exp = "({a}>{b}?{a}:{b})".format(a=a, b=b)
        return exp

    @staticmethod
    def min(a, b):
        """ Generates the expression min(a, b) to insert on GEE expression

        :param a: one value to compare
        :param b: the other value to compare
        :return: expression
        :rtype: str
        """

        exp = "({a}<{b}?{a}:{b})".format(a=a, b=b)
        return exp

    @staticmethod
    def parse(expr):
        s = SvalEE()
        return s.eval(expr)


def cat(op, group=True):
    def wrap(a, b):
        if group:
            return "({}{}{})".format(a, op, b)
        else:
            return "{}{}{}".format(a, op, b)
    return wrap


def cat_fun(nom_fun):
    def wrap(arg):
        return "{}({})".format(nom_fun, arg)
    return wrap


def cat_band(band):
    return "b('{}')".format(band)


# CLEAN sval
DEFAULT_NAMES = {"pi": math.pi,
                 "e": math.e}
DEFAULT_FUNCTIONS = {"max": ExpGen.max,
                     "min": ExpGen.min,
                     "b": cat_band,
                     "exp": cat_fun("exp"),
                     "sqrt": cat_fun("sqrt")}
DEFAULT_OPERATORS = {ast.Add: cat("+"),
                     ast.UAdd: lambda a: '+{}'.format(a),
                     ast.Sub: cat("-"),
                     ast.USub: lambda a: '-{}'.format(a),
                     ast.Mult: cat("*"),
                     ast.Div: cat("/"),
                     ast.FloorDiv: cat("//"),
                     ast.Pow: cat("**"),
                     ast.Mod: cat("%"),
                     ast.Eq: cat("=="),
                     ast.NotEq: cat("!="),
                     ast.Gt: cat(">"),
                     ast.Lt: cat("<"),
                     ast.GtE: cat(">="),
                     ast.LtE: cat("<=")}
                     # ast.Not: op.not_,
                     # ast.USub: op.neg,
                     # ast.UAdd: op.pos,
                     # ast.In: lambda x, y: op.contains(y, x),
                     # ast.NotIn: lambda x, y: not op.contains(y, x),
                     # ast.Is: lambda x, y: x is y,
                     # ast.IsNot: lambda x, y: x is not y,}


class SvalEE(sval.SimpleEval):
    def __init__(self, **kwargs):
        super(SvalEE, self).__init__(**kwargs)

        self.operators = DEFAULT_OPERATORS
        self.functions = DEFAULT_FUNCTIONS
        self.names = DEFAULT_NAMES


# -*- coding: utf-8 -*-
""" Tools to use in IPython """

import ee
from geetools import tools
from . import date, functions


def information():
    pass


def info2map(map):
    """ Add an information Tab to a map displayed with `geetools.ipymap`
    module

    :param map: the Map where the tab will be added
    :type map: geetools.ipymap.Map
    :return:
    """
    try:
        from ipywidgets import Accordion
    except:
        print('Cannot use ipytools without ipywidgets installed\n'
              'ipywidgets.readthedocs.io')

    map.addTab('BAP Inspector', info_handler, Accordion())


def info_handler(**kwargs):
    """ Handler for the Bap Inspector Tab of the Map

    :param change:
    :return:
    """
    try:
        from ipywidgets import HTML
    except:
        print('Cannot use ipytools without ipywidgets installed\n'
              'ipywidgets.readthedocs.io')

    themap = kwargs['map']
    widget = kwargs['widget']
    # Get click coordinates
    coords = kwargs['coordinates']

    event = kwargs['type'] # event type
    if event == 'click':  # If the user clicked
        # create a point where the user clicked
        point = ee.Geometry.Point(coords)

        # First Accordion row text (name)
        first = 'Point {} at {} zoom'.format(coords, themap.zoom)

        # Reset list of widgets and names
        namelist = [first]
        wids4acc = [HTML('')] # first row has no content

        length = len(themap.EELayers.keys())
        i = 1

        for name, obj in themap.EELayers.items(): # for every added layer
            # Clear children // Loading
            widget.children = [HTML('wait a second please..')]
            widget.set_title(0, 'Click on {}. Loading {} of {}'.format(coords, i, length))
            i += 1

            # IMAGES
            if obj['type'] == 'Image':
                # Get the image's values
                image = obj['object']
                properties = image.propertyNames().getInfo()

                if 'BAP_version' in properties:  # Check if it's a BAP composite
                    try:
                        values = tools.image.getValue(image, point, 10, 'client')
                        values = tools.dictionary.sort(values)
                        col_id = int(values['col_id'])
                        thedate = int(values['date'])
                        collection = functions.get_id_col(col_id)
                        realdate = date.Date.get(thedate).format().getInfo()

                        # Get properties of the composite
                        inidate = int(image.get('ini_date').getInfo())
                        inidate = date.Date.get(inidate).format().getInfo()
                        enddate = int(image.get('end_date').getInfo())
                        enddate = date.Date.get(enddate).format().getInfo()

                        # Create the content
                        img_html = '''
                            <h3>General Properties</h3>
                            <b>Season starts at:</b> {ini}</br>
                            <b>Season ends at:</b> {end}</br>
                            <h3>Information at point</h3>
                            <b>Collection:</b> {colid} ({col})</br>
                            <b>Date:</b> {thedate} ({date})'''.format(ini=inidate,
                            end=enddate, col=collection, date=realdate, p=coords,
                            thedate=thedate, colid=col_id)

                        wid = HTML(img_html)
                        # append widget to list of widgets
                        wids4acc.append(wid)
                        namelist.append(name)
                    except Exception as e:
                        widget = HTML(str(e).replace('<','{').replace('>','}'))
                        text = 'ERROR in layer {}'.format(name)
                        wids4acc.append(widget)
                        namelist.append(text)
                else:
                    continue
            # GEOMETRIES
            elif obj['type'] == 'Geometry':
                continue

        # BAP Widget
        # bapwid = m.childrenDict['BAP Inspector']
        widget.children = wids4acc

        for i, n in enumerate(namelist):
            widget.set_title(i, n)


# -*- coding: utf-8 -*-
import ee
from datetime import date
from geetools import collection
from geetools.collection.group import CollectionGroup

# IDS
ID1 = 'LANDSAT/LM01/C01/T1'
ID2 = 'LANDSAT/LM02/C01/T1'
ID3 = 'LANDSAT/LM03/C01/T1'
ID4TOA = 'LANDSAT/LT04/C01/T1_TOA'
ID4SR = 'LANDSAT/LT04/C01/T1_SR'
ID5TOA = 'LANDSAT/LT05/C01/T1_TOA'
ID5SR = 'LANDSAT/LT05/C01/T1_SR'
ID7TOA = 'LANDSAT/LE07/C01/T1_TOA'
ID7SR = 'LANDSAT/LE07/C01/T1_SR'
ID8TOA = 'LANDSAT/LC08/C01/T1_TOA'
ID8SR = 'LANDSAT/LC08/C01/T1_SR'
S2 = 'COPERNICUS/S2'
S2SR = 'COPERNICUS/S2_SR'


class SeasonPriority(object):
    """ Satellite priorities for seasons.

    :param breaks: list of years when there is a break
    :param periods: nested list of periods
    :param satlist: nested list of satellites in each period
    :param relation: dict of relations
    :param ee_relation: EE dict of relations
    """
    breaks = [1972, 1974, 1976, 1978, 1982, 1983,
              1994, 1999, 2003, 2012, 2013, date.today().year+1]

    periods = []
    for i, b in enumerate(breaks):
        if i < len(breaks) - 1:
            periods.append(range(b, breaks[i + 1]))

    satlist = [[ID1], # 72 74
               [ID2, ID1], # 74 76
               [ID3, ID2, ID1], # 76 78
               [ID3, ID2], # 78 82
               [ID4SR, ID4TOA, ID3, ID2], # 82 83
               [ID5SR, ID5TOA, ID4SR, ID4TOA], # 83 94
               [ID5SR, ID5TOA], # 94 99
               [ID7SR, ID7TOA, ID5SR, ID5TOA], # 99 03
               [ID5SR, ID5TOA, ID7SR, ID7TOA], # 03 12
               [ID8SR, ID8TOA, ID7SR, ID7TOA, ID5SR, ID5TOA], # 12 13
               [ID8SR, ID8TOA, ID7SR, ID7TOA]] # 13 -

    relation = dict(
        [(p, sat) for per, sat in zip(periods, satlist) for p in per])

    ee_relation = ee.Dictionary(relation)

    l7_slc_off = range(2003, date.today().year+1)

    def __init__(self, year):
        self.year = year

    @property
    def satellites(self):
        '''
        :return: list of satellite's ids
        :rtype: list
        '''
        return self.relation[self.year]

    @property
    def collections(self):
        '''
        :return: list of satcol.Collection
        :rtype: list
        '''
        sat = self.satellites
        return [collection.fromId(id) for id in sat]

    @property
    def colgroup(self):
        '''
        :rtype: CollectionGroup
        '''
        return CollectionGroup(*self.collections)
